#include "btree.h"
#include "kine.h"

bbTNode::bbTNode() {

  lson = NULL; rson = NULL; parent = NULL;
  mindes = NULL; point = NULL; tree = NULL;
  color = RED; rank = 0; minskey = MAX;
}

bbTNode::bbTNode(bbTNode* p) {

  lson = NULL; rson = NULL; parent = p;
  mindes = NULL; point = NULL; tree = NULL;
  color = RED; rank = 0; minskey = MAX;
}

bbTNode::bbTNode(mObj2D* p, short axis) {

  makeNode(p,axis);
}

bbTNode::bbTNode(mObj2D* p, short axis, bbTNode* pp) {

  makeNode(p,axis,pp);
}

void
bbTNode::makeNode(mObj2D* p, short axis) {

  lson = NULL; rson = NULL; parent = NULL;
  mindes = this; point = p; tree = NULL;
  color = RED; rank = 0; minskey = p->ord(axis);
}

void  
bbTNode::makeNode(mObj2D* p, short axis, bbTNode* pp) {

  makeNode(p,axis);
  parent = pp;
}

void
bbTNode::reset() {

  lson = NULL; rson = NULL; parent = NULL;
  mindes = this; tree = NULL; color = RED; 
  rank = 0;
}

void
bbTNode::reset(short axis) {

  lson = NULL; rson = NULL; parent = NULL;
  mindes = this; tree = NULL;
  color = RED; rank = 0; 
  if(point!=NULL) minskey = point->ord(axis);
}

void
bbTNode::deleteNode(void) {

  if(this==NULL) return;
  if(lson!=NULL) {
    lson->deleteNode();
    lson = NULL;
  }
  if(rson!=NULL) {
    rson->deleteNode();
    rson = NULL;
  }
  delete this;
} 

void 
bbTNode::printValue(void) {

  point->draw(0);
}

void
bbTNode::printIn(void) {

  if(lson!=NULL) lson->printIn();
  printValue();
  if(rson!=NULL) rson->printIn();
}

void
bbTNode::printPost(void) {

  if(lson!=NULL) lson->printPost();
  if(rson!=NULL) rson->printPost();
  printValue();
}

void
bbTNode::printPre(void) {

  printValue();
  if(lson!=NULL) lson->printPre();
  if(rson!=NULL) rson->printPre();
}

void 
bbTNode::printValue(short axis) {

  point->ordaxis[axis].draw(0);
}

void
bbTNode::printIn(short axis) {

  if(lson!=NULL) lson->printIn(axis);
  printValue(axis);
  if(rson!=NULL) rson->printIn(axis);
}

void
bbTNode::printPost(short axis) {

  if(lson!=NULL) lson->printPost(axis);
  if(rson!=NULL) rson->printPost(axis);
  printValue(axis);
}

void
bbTNode::printPre(short axis) {

  printValue(axis);
  if(lson!=NULL) lson->printPre(axis);
  if(rson!=NULL) rson->printPre(axis);
}

void
bbTNode::updateup(short axis) {

  bbTNode* curr;
  curr = this;
  while(curr!=NULL) {
    curr->update(axis);
    curr = curr->parent;
  }
}

void
bbTNode::update(short axis) {

  mindes = this;
  minskey = point->ord(axis);
  if(lson!=NULL) {
    if(lson->minskey<minskey) {
      minskey = lson->minskey;
      mindes = lson->mindes;
    }       
  }
  if(rson!=NULL) {
    if(rson->minskey<minskey) {
      minskey = rson->minskey;
      mindes = rson->mindes;
    } 
  }
}

bbTNode *
bbTNode::root(void) {

  bbTNode *pa;
  pa = this;
  while(pa->parent!=NULL) pa = pa->parent;
  return pa;
}

bbTree *
bbTNode::relTree(void) {

  bbTNode *r;
  r = root();
  if(r==NULL) return NULL;
  return r->tree;
}

bbTNode* 
bbTNode::rotate(short dir, short axis) {

  bbTNode* temp;
  if(dir == RIGHT) {

    if(lson==NULL) return NULL;
    
    temp = lson->rson;
    lson->rson = this;
    lson->parent = parent;
    if(isroot()) {
      tree->root = lson;
      lson->tree = tree;
    } else if(islson())
      parent->lson = lson;
    else
      parent->rson = lson;    
    parent = lson;
    lson = temp;
    if(temp!=NULL) temp->parent = this;
    parent->mindes = mindes;
    parent->minskey = minskey;
    update(axis);
    return parent;
  } else if(dir == LEFT) {

    if(rson==NULL) return NULL;
    temp = rson->lson;
    rson->lson = this;
    rson->parent = parent;
    if(isroot()) {
      tree->root = rson;
      rson->tree = tree;
    } else if(islson())
      parent->lson = rson;
    else
      parent->rson = rson;    
    parent = rson;
    rson = temp;
    if(temp!=NULL) temp->parent = this;
    parent->mindes = mindes;
    parent->minskey = minskey;
    update(axis);
    return parent;
  }
  return NULL;
}

bbTNode::verify(void) {

  if(parent==NULL) 
    return (tree==NULL)?1:(tree->root == this);
  return (parent->lson==this)||(parent->rson==this);
}

void
bbTNode::decompose(short axis) {

  if(lson!=NULL) lson->decompose(axis);
  if(rson!=NULL) rson->decompose(axis);
  lson = NULL; rson = NULL; parent = NULL;
  mindes = this; tree = NULL;
  color = RED; rank = 0; minskey = point->ord(axis);
}

short
bbTNode::compare(bbTNode *b) {

  if(b==NULL) return 0;
  if(point!=b->point) return 0;
  if(((lson!=NULL)&&(b->lson==NULL))||
     ((lson==NULL)&&(b->lson!=NULL)))
    return 0;
  if((lson!=NULL)&&(b->lson!=NULL)) {
    if(!lson->compare(b->lson)) return 0;
  } 
  if(((rson!=NULL)&&(b->rson==NULL))||
     ((rson==NULL)&&(b->rson!=NULL)))
    return 0;
  if((rson!=NULL)&&(b->rson!=NULL)) {
    if(!rson->compare(b->rson)) return 0;
  } 
  return 1;
}


bbTree::bbTree(void) { 

  root = NULL; ax = X;
}

bbTree::bbTree(short a) { 

  root = NULL; ax = a;
}

bbTree::bbTree(bbTNode* r)  {
  
  root = r; ax = X; 
  if(root!=NULL) {
    root->tree = this;
    root->color = BLACK;
  }
}

bbTree::bbTree(bbTNode* r, short a)  {

  root = r; ax = a; 
  if(root!=NULL) {
    root->tree = this;
    root->color = BLACK;
  }
}

bbTree::bbTree(mObj2D* rl, short a) {

  root = NULL; ax = a; relmo = rl;
}

bbTree::bbTree(mObj2D* rl, bbTNode* r) {

  root = r; ax = X; 
  if(root!=NULL) {
    root->tree = this;
    root->color = BLACK;
  }
  relmo=rl;
}

bbTree::bbTree(mObj2D* rl, bbTNode* r, short a)  {
  
  root = r; ax = a; 
  if(root!=NULL) {
    root->tree = this;
    root->color = BLACK;
  }
  relmo = rl;
}

void
bbTree::makeTree(bbTNode *r, short a) {

  root = r; ax = a; 
  if(root!=NULL) {
    root->tree = this;
    root->color = BLACK;
  }
}

void
bbTree::makeTree(mObj2D *rl, short a) {

  root = NULL; ax = a; relmo = rl;
}

void
bbTree::makeTree(mObj2D *rl, bbTNode *r, short a) {

  root = r; ax = a; 
  if(root!=NULL) {
    root->tree = this;
    root->color = BLACK;
  }
  relmo = rl;
}

void
bbTree::setrelmo(mObj2D *rl) {
  
  relmo = rl;
}

bbTree* 
bbTree::cut(bbTNode *r) {

  if(r==NULL) return NULL;
  if(r->isroot()) {
    root = NULL;
  } else {
    if(r->islson())
      r->parent->lson = NULL;
    else
      r->parent->rson = NULL;
    r->parent->updateup(ax);
    r->parent = NULL;
  }
  return(new bbTree(r,ax));
}


void
bbTree::deleteTree() {

  if(this==NULL) return;
  if(root==NULL) return;
  root->deleteNode();  
}

int 
bbTree::bheight() {

  bbTNode *curr;
  int bh = 0;

  curr = root;
  while(curr!=NULL) {
    if(curr->color == BLACK) bh++;
    curr = curr->lson;
  }
  return bh;
}

bbTNode*
bbTree::search(int key, short axis, short *pdir) {

  bbTNode *current, *old;
  int key2;

  if(root==NULL) return NULL;
  current = root;
  while(current!=NULL) {

    old = current;
    key2 = current->point->ord(axis);
    if(key<=key2)
      current = current->lson;
    else
      current = current->rson;
  }
  
  if(key<=key2)
    *pdir = LEFT;
  else
    *pdir = RIGHT;

  return old;
}

  
bbTNode*
bbTree::search(mObj2D* p, short axis, short *pdir) {

  int key = p->ord(axis);
  return(search(key,axis,pdir));
}

void
bbTree::insfix(bbTNode* nn) {

  bbTNode *x = nn, *y;

  x->color = RED;
  while(!x->isroot()&&(x->parent->color == RED)) {
    if(x->parent->islson()) {
      y = x->parent->parent->rson;
      if(Color(y) == RED) {
	x->parent->color = BLACK;
	y->color = BLACK;
	x->parent->parent->color = RED;
	x = x->parent->parent;
      } else {
	if(x->isrson()) {
	  x = x->parent;
	  x->rotate(LEFT,ax);
	}
	x->parent->color = BLACK;
	x->parent->parent->color = RED;
	x->parent->parent->rotate(RIGHT,ax);
      }
    } else { 
      y = x->parent->parent->lson;
      if(Color(y) == RED) {
	x->parent->color = BLACK;
	y->color = BLACK;
	x->parent->parent->color = RED;
	x = x->parent->parent;
      } else {
	if(x->islson()) {
	  x = x->parent;
	  x->rotate(RIGHT,ax);
	}
	x->parent->color = BLACK;
	x->parent->parent->color = RED;
	x->parent->parent->rotate(LEFT,ax);
      }
    }
  }
  root->color = BLACK;
}

bbTNode*
bbTree::insert(bbTNode* nn, short axis) {

  if(tinsert(nn,axis)==NULL) return NULL;
  insfix(nn);
  return nn;
}
      
bbTNode*
bbTree::tinsert(bbTNode* nn,short axis) {

  if(nn==NULL) return NULL;
  if(root==NULL) {
    root = nn;
    root->tree = this;
    return nn;
  }

  bbTNode *current, *old;
  int key1,key2,skey;

  key1 = nn->point->ord(axis);
  skey = nn->minskey;

  old = NULL;
  current = root;

  while(current!=NULL) {

    old = current;
    if(skey<current->minskey) {
      current->minskey = skey;
      current->mindes = nn;
    }

    key2 = current->point->ord(axis);
    if(key1<key2)
      current = current->lson;
    else
      current = current->rson;
  }

  if(key1<key2)
    old->lson = nn;
  else 
    old->rson = nn;
  nn->parent = old;
  root->tree = this;
  return nn;
}

void
bbTree::delfix(bbTNode *fixp, bbTNode *cutp, short dir) {

  bbTNode *w,*nil,*x;

  if(fixp!=NULL) {
    x = fixp;
    nil = NULL;
  } else {
    if(cutp==NULL) return;
    nil = new bbTNode;
    nil->parent = cutp;
    if(dir==LEFT) cutp->lson = nil;
    else cutp->rson = nil;
    nil->color = BLACK;
    x = nil;
  } 

  while(!x->isroot()&&(x->color == BLACK)) {
    if(x->islson()) {
      w = x->parent->rson;
      if(w==NULL) break;
      if(Color(w) == RED) {
	w->color = BLACK;
	x->parent->color = RED;
	x->parent->rotate(LEFT,ax);
	w = x->parent->rson;
      }
      if(w==NULL) break;
      if((Color(w->lson)==BLACK)&&(Color(w->rson)==BLACK)) {
	w->color = RED;
	x = x->parent;
      } else {
	if(Color(w->rson) == BLACK) {
	  w->lson->color = BLACK;
	  w->color = RED;
	  w->rotate(RIGHT,ax);
	  w = x->parent->rson;
	}
	w->color = x->parent->color;
	x->parent->color = BLACK;
	w->rson->color = BLACK;
	x->parent->rotate(LEFT,ax);
	x = root;
      }
    } else {

      w = x->parent->lson;
      if(w==NULL) break;
      if(Color(w) == RED) {
	w->color = BLACK;
	x->parent->color = RED;
	x->parent->rotate(RIGHT,ax);
	w = x->parent->lson;
      }
      if(w==NULL) break;
      if((Color(w->lson)==BLACK)&&(Color(w->rson)==BLACK)) {
	w->color = RED;
	x = x->parent;
      } else {
	if(Color(w->lson)==BLACK) {
	  w->rson->color = BLACK;
	  w->color = RED;
	  w->rotate(LEFT,ax);
	  w = x->parent->lson;
	}
	w->color = x->parent->color;
	x->parent->color = BLACK;
	w->lson->color = BLACK;
	x->parent->rotate(RIGHT,ax);
	x = root;
      }
    }
  }
  if(root!=NULL) root->color = BLACK;
  x->color = BLACK;
  if(nil!=NULL) {
    if(nil->islson()) nil->parent->lson = NULL;
    else nil->parent->rson = NULL;
    delete nil;
  }
}

bbTNode*
bbTree::del(bbTNode* onode) {

  bbTNode *x,*y;
  short c,dir;
  x = tdel(onode,&c,&y,&dir);
  if(c==RED) return x;
  delfix(x,y,dir);
  return x;
}
    
bbTNode*
bbTree::tdel(bbTNode* onode, short* c, bbTNode** pnode, short* pdir) {

  if(onode==NULL) return NULL;
  if(onode->islson()) *pdir = LEFT;
  else *pdir = RIGHT;
  if(onode->isleaf()) {

    if(onode->isroot())
      root = NULL;
    else {
      if(onode->islson())
	onode->parent->lson = NULL;
      else
	onode->parent->rson = NULL;
      onode->parent->updateup(ax);
    }
    *c = onode->color;
    *pnode = onode->parent;
    return NULL;
  }

  if(onode->lson == NULL) {

    if(onode->isroot()) {
      onode->rson->parent = NULL;
      root = onode->rson;
      root->tree = this;
    } else {
      if(onode->islson()) {
        onode->parent->lson = onode->rson;
        onode->rson->parent = onode->parent;
      } else {
        onode->parent->rson = onode->rson;
        onode->rson->parent = onode->parent;
      } 
      onode->parent->updateup(ax);
    }
    *c = onode->color;
    *pnode = onode->parent;
    return onode->rson;
  }

  if(onode->rson == NULL) {

    if(onode->isroot()) {
      onode->lson->parent = NULL;
      root = onode->lson;
      root->tree = this;
    } else { 
      if(onode->islson()) {
        onode->parent->lson = onode->lson;
        onode->lson->parent = onode->parent;
      } else {
        onode->parent->rson = onode->lson;
        onode->lson->parent = onode->parent;
      }
      onode->parent->updateup(ax);
    }
    *c = onode->color;
    *pnode = onode->parent;
    return onode->lson;
  }

  // complicated case - having two children

  // find split node

  bbTNode *curr,*cutp,*fixp;
  curr = onode->rson;  
  while(curr->lson!=NULL) curr = curr->lson;

  if(curr->islson()) *pdir = LEFT;
  else *pdir = RIGHT;

  cutp = curr->parent;
  fixp = curr->rson;

  if(curr->islson()) curr->parent->lson = curr->rson;
  else curr->parent->rson = curr->rson;

  if(curr->rson!=NULL) curr->rson->parent = curr->parent;
    
  curr->lson = onode->lson;
  curr->rson = onode->rson;
  onode->lson->parent = curr;
  if(onode->rson!=NULL)
    onode->rson->parent = curr;
  
  *c = curr->color;
  curr->color = onode->color;

  if(onode->isroot()) {
    root = curr;
    root->parent = NULL;
    root->tree = this;
  } else {
    if(onode->islson())
      onode->parent->lson = curr;
    else // onode is rson
      onode->parent->rson = curr;
    curr->parent = onode->parent;
  }
  if(cutp==onode) curr->updateup(ax);
  else cutp->updateup(ax);
  *pnode = fixp;
  return cutp;
}

void 
bbTree::ljoinwith(bbTree* bt) {

  if(bt==NULL) return;
  if(bt->root == NULL) return;
  bbTNode *cutp;
  cutp = bt->root;
  while(cutp->lson!=NULL) cutp = cutp->lson;
  bt->del(cutp);
  cutp->reset(ax);
  ljoinwith(bt,cutp);
}

void
bbTree::ljoinwith(bbTree* bt, bbTNode* bn) {
 
  bbTree tbt(ax);

  tbt.join(this,bt,bn);
  root = tbt.root;
  root->tree = this;
}

void 
bbTree::rjoinwith(bbTree* bt) {

  if(bt==NULL) return;
  if(bt->root == NULL) return;
  bbTNode *cutp;
  cutp = bt->root;
  while(cutp->rson!=NULL) cutp = cutp->rson;
  bt->del(cutp);
  cutp->reset(ax);
  rjoinwith(bt,cutp);
}

void
bbTree::rjoinwith(bbTree* bt, bbTNode* bn) {

  bbTree tbt(ax);

  tbt.join(bt,this,bn);
  root = tbt.root;
  root->tree = this;
}

void 
bbTree::join(bbTree* bt1, bbTree* bt2, bbTNode* bn) {

  if(bn==NULL) return;

  int bh1, bh2,bh;
  bbTNode *insp,*inspp,*bt1root,*bt2root;

  ax = X;

  if(bt1==NULL) {
    bh1 = 0;
    bt1root = NULL;
  } else {
    ax = bt1->ax;
    bh1 = bt1->bheight();
    bt1root = bt1->root;
  }

 if(bt2==NULL) {
    bh2 = 0;
    bt2root = NULL;
  } else {
    ax = bt2->ax;
    bh2 = bt2->bheight();
    bt2root = bt2->root;
  }

  if((bh1==0)&&(bh2==0)) {

    root = bn;
    bn->color = BLACK;
    root->tree = this;

  } else if(bh1==bh2) {

    bn->color = BLACK;
    bn->lson = bt1root;
    bn->rson = bt2root;
    bt1root->parent = bn;
    bt2root->parent = bn;
    bn->update(ax);
    root = bn;
    root->tree = this;

  } else if(bh1<bh2) {

    insp = bt2root;
    bh = bh2;
    while(insp!=NULL) {
      if(insp->color==BLACK) bh--;
      if(bh<bh1) break;
      inspp = insp;
      insp = insp->lson;
    }
    bn->color = RED;
    inspp->lson = bn;
    bn->parent = inspp;
    bn->lson = bt1root;
    bn->rson = insp;
    if(bt1root!=NULL) bt1root->parent = bn;
    if(insp!=NULL) insp->parent = bn;
    bn->updateup(ax);
    root = bt2root;
    root->tree = this;
    insfix(bn);
  } else if(bh2<bh1) {
    insp = bt1root;
    bh = bh1;
    while(insp!=NULL) {
      if(insp->color==BLACK) bh--;
      if(bh<bh2) break;
      inspp = insp;
      insp = insp->rson;
    }

    bn->color = RED;
    inspp->rson = bn;
    bn->parent = inspp;
    bn->lson = insp;
    bn->rson = bt2root;
    if(bt2root!=NULL) bt2root->parent = bn;
    if(insp!=NULL) insp->parent = bn;
    bn->updateup(ax);
    root = bt1root;
    root->tree = this;
    insfix(bn);
  }
}


void
bbTree::split(bbTree* bt, bbTNode* bn, short dir, short l) {

  bbTree *btt;
  btt = new bbTree;
  if(l==0)
    split(btt,bt,bn,dir);
  else
    split(bt,btt,bn,dir);
  root = btt->root;
  if(root!=NULL) root->tree = this;
  delete btt;
}

void
bbTree::split(bbTree* bt1, bbTree* bt2, mObj2D* p, short axis) {

  bbTNode* a;
  short dir;

  a = search(p,axis,&dir);
  split(bt1,bt2,a,dir);
}

void
bbTree::split(bbTree* bt1, bbTree*bt2, int s, short axis) {
  
  bbTNode* a;
  short dir;
  a = search(s,axis,&dir);
  split(bt1, bt2, a, dir);
}

void
bbTree::split(bbTree* bt, mObj2D* p, short axis, short l) {

  bbTNode* a;
  short dir;

  a = search(p,axis,&dir);
  split(bt,a,dir,l);
}

void
bbTree::split(bbTree* bt, int s, short axis, short l) {
  
  bbTNode* a;
  short dir;
  a = search(s,axis,&dir);
  split(bt,a,dir,l);
}
    
void
bbTree::split(bbTree* bt1, bbTree* bt2, bbTNode* bn, short dir) {

  // bn -> bt1
  bt1->ax = ax;
  bt2->ax = ax;
  bt1->root = NULL;
  bt2->root = NULL;

  if(root==NULL) return;
  if(bn == NULL) return; 
  bbTree *br1, *br2, *btemp;
  bbTNode *curr,*old;
  short flag;
  curr = bn->parent;
  if(!bn->isroot()) {
    if(bn->islson()) flag = LEFT;
    else flag = RIGHT;
  }  
  if(dir==RIGHT) { 
    btemp = cut(bn);
    br1 = btemp->cut(bn->lson);
    bn->reset(ax);
    bt1->rjoinwith(br1,bn);
  } else if(dir==LEFT) {
    btemp = cut(bn);
    br2 = btemp->cut(bn->rson);
    bn->reset(ax);
    bt2->ljoinwith(br2,bn);
  }

  while(curr!=NULL) {
    old = curr;
    curr = curr->parent;
    if(flag==LEFT) {
      if(old->islson()) flag = LEFT;
      else flag = RIGHT;
      btemp = cut(old);
      br2 = btemp->cut(old->rson);
      old->reset(ax);
      bt2->ljoinwith(br2,old);
    } else {
      if(old->islson()) flag = LEFT;
      else flag = RIGHT;
      btemp = cut(old);
      br1 = btemp->cut(old->lson);
      old->reset(ax);
      bt1->rjoinwith(br1,old);
    }
  }
  if(bt1->root!=NULL) bt1->root->tree = bt1;
  if(bt2->root!=NULL) bt2->root->tree = bt2;
}

bbTNode*
bbTree::minSkey(void) {

  if(root==NULL) return NULL;
  return root->mindes;
}

bbTNode*
bbTree::max(void) {

  bbTNode *curr;
  if(root==NULL) return NULL;
  curr = root;
  while(curr->rson!=NULL) curr = curr->rson;
  return curr;
}

bbTNode*
bbTree::min(void) {

  bbTNode *curr;
  if(root==NULL) return NULL;
  curr = root;
  while(curr->lson!=NULL) curr = curr->lson;
  return curr;
}

void
bbTree::decompose(void) {

  if(root!=NULL) root->decompose(ax);
}

void 
bbTree::printTreeIn(void) {

  if(root==NULL) cout<<"****NULL****\n";
  else root->printIn();
}

void 
bbTree::printTreePost(void) {

  if(root==NULL) cout<<"****NULL****\n";
  else root->printPost();
}

void 
bbTree::printTreePre(void) {

  if(root==NULL) cout<<"****NULL****\n";
  else root->printPre();
}

void
bbTree::printTreeIn(short axis) {

  if(root==NULL) cout<<"****NULL****\n";
  else root->printIn(axis);
}

void 
bbTree::printTreePost(short axis) {

  if(root==NULL) cout<<"****NULL****\n";
  else root->printPost(axis);
}

void 
bbTree::printTreePre(short axis) {

  if(root==NULL) cout<<"****NULL****\n";
  else root->printPre(axis);
}

short
bbTree::compare(bbTree *bt) {

  if(root==NULL) {
    if(bt->root==NULL) return 1;
    else return 0;
  }
  return root->compare(bt->root);
}
    





