
#include "draw.h"
#include "btree.h"
#include "kine.h"
#include "trackball.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/times.h>

GLfloat ambient[4] = { 0.3, 0.3, 0.3, 1.0 };
GLfloat diffuse[4] = { 1.0, 1.0, 1.0, 1.0 };
GLfloat specular[4] = { 1.0, 1.0, 1.0, 1.0 };
GLfloat position[4] = { 20.0, 30.0, 10.0, 0.0 };
GLfloat lmodel_ambient[4] = { 0.4, 0.4, 0.4, 1.0 };
GLfloat local_view[1] = { 0.0 };

GLfloat no_mat[4] = { 0.0, 0.0, 0.0, 1.0 };
GLfloat mat_ambient[4] = { 0.7, 0.7, 0.7, 1.0 };
GLfloat mat_ambient_color[4] = { 0.8, 0.2, 0.2, 1.0 };
GLfloat mat_diffuse[4] = { 0.7, 0.5, 0.3, 1.0 };
GLfloat mat_specular[4] = { 1.0, 1.0, 1.0, 1.0 };
GLfloat no_shininess[1] = { 0.0 };
GLfloat low_shininess[1] = { 5.0 };
GLfloat high_shininess[1] = { 100.0 };
GLfloat mat_emission[4] = {0.3, 0.2, 0.2, 0.0};

struct timeval starttime,gt;
mObjs *mos;
mObj2D *o1 = NULL, *o2 = NULL;
REAL critT, nextT, startT, currT;
short process;

extern FD_drawingwindow *fdraw;

float oldYaw = 0, newYaw = 0, Yaw = 0;
float oldPitch = 0, newPitch = 0, Pitch = 0;

void 
beginRotate() {

}

void
doRotate() {

}

void 
endRotate() {
}
void
resetTimer(void)
{
  gettimeofday(&starttime);
}

void
resetGTimer(void)
{
  gettimeofday(&gt);
}

float
readGTimer(void) {

  struct timeval currtime;
  gettimeofday(&currtime);
  return ((currtime.tv_usec - gt.tv_usec) * 1.0e-4 +
	  (currtime.tv_sec - gt.tv_sec)*100.0);
}

float
readTimer(void)
{
    struct timeval currtime;
    gettimeofday(&currtime);
    return ((currtime.tv_usec - starttime.tv_usec) * 1.0e-6 +
            (currtime.tv_sec - starttime.tv_sec));
}

void initRandom(void)
{
  int stime;
  long ltime;
  ltime = time(NULL);
  stime = (unsigned) ltime/2;
  srand(stime);
  cout<<"*******  "<<stime<<"  *******\n";
}

void initRandom(int seed) {

  if(seed==0) initRandom();
  else srand(seed);
}

void move(void) {

  process = 1;
  mos->processevent(&o1,&o2,&critT,&nextT);
  resetGTimer();
  startT = currT;
  process = 0;
}

void glinit()
{
  glDepthFunc(GL_LESS);
  glEnable(GL_DEPTH_TEST);
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
  glLightfv(GL_LIGHT0, GL_POSITION, position);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  //glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, local_view);
  
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess);
  glMaterialfv(GL_FRONT, GL_EMISSION, no_mat);
  
  glClearColor(0.0, 0.0, 0.0, 0.0);
}

void initparameters() {

  initRandom();
  mos = new mObjs(10,3.0);
  mos->setboundary(13,87,13,87);
  if(!mos->gen_nocollision(0.0,0.1,100)) {
    
    printf("Can not generate initial configuration\n");
    printf("Please adjust your parameter - boundary, number of balls, radius");
    exit(0);
  }
  critT = 0; nextT = 0; startT = 0; currT = 0;  
  mos->getcpair(&o1,&o2,-1,0);
  nextT = mos->nextcritical();
  resetGTimer();
  resetTimer();

  process = 0;
}

void 
setview() {

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(W/2,L/2,0);
  glRotatef(Yaw/20.0,0,1,0);
  glRotatef(Pitch/20,1,0,0);
  glTranslatef(-W/2,-L/2,0);
}


int
drawindow(XEvent *ev, void *data) {

  if(process) return 0;
  setview();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
  mos->drawBoundary();

  currT = readGTimer()+startT;  
  fl_set_idle_callback(0,0);
  while (currT>=nextT) move();

  for(int i=0;i<mos->nofObjs;i++)
    mos->mos[i].draw(currT);
  glFinish();
  glXSwapBuffers(fl_display,fl_get_canvas_id(fdraw->drawindow));
  fl_set_idle_callback(drawindow,0);
  return 0;
}

int
canvas_expose(FL_OBJECT *ob, Window win, int w, int h, XEvent *xev, void *ud) {
  
  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if (w < h) 
    glOrtho (0.0, 100.0, 0.0, 100.0*(GLfloat) h/(GLfloat) w, -100.0, 100.0);
  else
    glOrtho (0.0, 100.0*(GLfloat) w/(GLfloat) h, 0.0, 100.0, -100.0, 100.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  drawindow(0,0);
  return 0;
}

int
begin_rotate(FL_OBJECT *ob, Window win, int w, int h, XEvent *xe, void *ud) {
   
  oldYaw = xe->xbutton.x;
  oldPitch = xe->xbutton.y;
  fl_add_canvas_handler(fdraw->drawindow, MotionNotify, do_rotate, 0);
  return 0;
}

int
do_rotate(FL_OBJECT *ob, Window win, int w, int h, XEvent *xe, void *ud) {

  newYaw = xe->xmotion.x-oldYaw;
  oldYaw = newYaw;
  Yaw += newYaw;
  newPitch= xe->xmotion.y-oldPitch;
  oldPitch = newPitch;
  Pitch += newPitch;
  return 0;
}

int
end_rotate(FL_OBJECT *ob, Window win, int w, int h, XEvent *xe, void *ud) {

  fl_remove_canvas_handler(fdraw->drawindow,MotionNotify,do_rotate);
  newYaw = xe->xbutton.x-oldYaw;
  oldYaw = newYaw;
  Yaw += newYaw;
  newYaw = 0;
  newPitch = xe->xbutton.y-oldPitch;
  oldPitch = newPitch;
  Pitch += newPitch;
  newPitch = 0;
  return 0;
}
