#include "kine.h"

short ver;

#define VE ver=V(this);

GLUquadricObj* ball;

void createBall() {

  ball = gluNewQuadric();
  gluQuadricNormals(ball, GLU_SMOOTH);
  gluQuadricDrawStyle(ball, GLU_FILL);
}

V(mObj2D * mo) {

  for(short axis=0;axis<=2;axis++) {

    if(!mo->relNode[axis].verify()) return 1;
    if(!mo->cuNode[axis].verify()) return 2;
    if(!mo->clNode[axis].verify()) return 3;
  }
  return 0;
}

V(mObjs * mos) {
  
  int v;
  for(int i = 0; i<mos->nofObjs; i++) 
    if( (v=V(&(mos->mos[i])))!=0) return v;
  return 0;
}

mObj1D::mObj1D(REAL sp, REAL ss) {
  
  startpoint = sp;
  speed = ss;
  starttime = 0;
}

mObj1D::mObj1D(REAL sp, REAL ss, REAL st) {
  
  startpoint = sp;
  speed = ss;
  starttime = st;
}

mObj1D::mObj1D(mObj2D* mo) {

  starttime = 0;
  org = mo;
}

void
mObj1D::init(mObj2D* mo) {

  starttime = 0;
  org = mo;
}

void	
mObj1D::set(REAL sp, REAL ss, REAL st) {
  
  startpoint = sp;
  speed = ss;
  starttime = st;
}

mObj1D::update(REAL sp, REAL ss, REAL st) {
  
  if(starttime>st) return 0;
  startpoint = sp;
  speed = ss;
  starttime = st;
  return 1;
}

mObj1D::update(mObj1D mo) {
  
  return update(mo.startpoint,mo.speed,mo.starttime);
}

REAL 
mObj1D::meet(mObj1D mo,REAL t) {
  
  if(speed >= mo.speed) return MAX;
  REAL mp = mo.startpoint+(starttime-mo.starttime)*mo.speed;
  REAL mt = (startpoint-mp)/(mo.speed-speed)+starttime;
  if(mt<starttime) return MAX;
  if(mt<mo.starttime) return MAX;
  if(mt<=t) return MAX;
  return mt;
}

REAL
mObj1D::distance(mObj1D mo, REAL t) {

  if(t<starttime) return -1;
  if(t<mo.starttime) return -1;
  REAL d = pos(t) - mo.pos(t);
  return(abs(d));
}

short
mObj1D::order(mObj1D* mo, REAL t) {
  
  if(t<starttime) return -2;
  if(t<mo->starttime) return -2;

  REAL currp1,currp2;
  
  currp1 = pos(t);
  currp2 = mo->pos(t);
  if(currp1==currp2) return 0;
  else if(currp1<currp2) return -1;
  else return 1;
}

void
mObj1D::draw(REAL t) {

  if(t<starttime) 
    cout<<"\n Illegal time \n";
  else
    cout<<"\n"<<pos(t)<<"\n";
}

flightPlan2::flightPlan2(P2 sp, P2 ss) {
  
  assign(sp,startpoint);
  assign(ss,speed);
  starttime = 0;
}

flightPlan2::flightPlan2(P2 sp, P2 ss, REAL st) {
  
  assign(sp,startpoint);
  assign(ss,speed);
  starttime = st;
}

flightPlan2::update(P2 sp, P2 ss, REAL st) {
  
  if(starttime>st) return 0;
  assign(sp,startpoint);
  assign(ss,speed);
  starttime = st;
  return 1;
}

flightPlan2::update(flightPlan2 fp) {
  
  return(update(fp.startpoint,fp.speed,fp.starttime));
}

void
flightPlan2::makeorder(mObj1D* pfp, short axis) {

  P2 p1,p2;
  rotate(startpoint,p1,axis);
  rotate(speed,p2,axis);
  pfp->set(p1[X],p2[X],starttime);
}

short
flightPlan2::pos(P2 p, REAL t) {

  if(t<starttime) return 0;
  
  p[X] = startpoint[X] + speed[X]*(t-starttime);
  p[Y] = startpoint[Y] + speed[Y]*(t-starttime);

  return 1;
}

flightPlan2::bounced(REAL bound, REAL t, short side) {

  P2 p;
  if(!pos(p,t)) return 0;
  if(side == X) {
    startpoint[X] = bound;
    startpoint[Y] = p[Y];
    speed[X] = -speed[X];
    starttime = t;
  } else if(side == Y) { 
    startpoint[X] = p[X];
    startpoint[Y] = bound;
    speed[Y] = -speed[Y];
    starttime = t;
  } else return 0;
  return 1;     
}

REAL
flightPlan2::hitBoundary(Boundary *b, REAL t, short *side) {

  REAL t1, t2;
  short f1,f2;
  short s;

  if(t<starttime) return MAX;
  if(speed[X]>0) {
    t1 = (b->right-startpoint[X])/speed[X]+starttime;
    f1 = 1;
  } else {
    t1 = (b->left-startpoint[X])/speed[X]+starttime;
    f1 = 0;
  }

  if(speed[Y]>0) {
    t2 = (b->top-startpoint[Y])/speed[Y]+starttime;
    f2 = 0;
  } else {
    t2 = (b->bottom-startpoint[Y])/speed[Y]+starttime;
    f2 = 1;
  }

  s = 2;
  if(t1>t) {
    if(t2>t) {
      if(t1>t2) s = Y ;
      else s = X;
    } else s = X;
  } else if(t2>t) s = Y;

  if(s==X){
    *side = 2*f1;
    return t1;
  } else if(s==Y) {
    *side = 2*f2+1;
    return t2;
  } else return MAX;
}

mObj2D::mObj2D(void) {

  for(short i=X;i<=X120;i++) { 
    ordaxis[i].init(this);
    lcand[i] = NULL;
    cand[i].makeTree(this, i);
    parentU[i].makeTree(this, i);
    parentL[i].makeTree(this, i);
  }
  radius = 0;
  thetastep = 0;
}


mObj2D::mObj2D(flightPlan2 fp2, REAL r) {

  mObj2D();
  radius = r;
  fp.update(fp2);
  for(short axis=X;axis<=X120;axis++)
    fp.makeorder(&(ordaxis[axis]),axis);
  thetastep = 0;
}

mObj2D::mObj2D(flightPlan2 fp2) {

  mObj2D(fp2,0);
}

mObj2D::mObj2D(P2 sp,P2 ss, REAL r) {

  mObj2D();
  set(sp,ss,0,r);
}

mObj2D::mObj2D(P2 sp, P2 ss) {

  mObj2D(sp,ss,0);
}

void
mObj2D::set(P2 sp, P2 ss, REAL t, REAL r) {

  radius = r;
  fp.update(sp,ss,t);
  for(short axis = X;axis<=X120;axis++) {
    fp.makeorder(&(ordaxis[axis]),axis);
    lcand[axis] = NULL;
  }
  normvec[X] = -ss[Y];
  normvec[Y] = ss[X];
  thetastep = norm(normvec)*10;
  normalize(normvec);
}

mObj2D::update(flightPlan2 fp2) {

  if(!fp.update(fp2)) return 0;
  fp.makeorder(&(ordaxis[X]),X);
  fp.makeorder(&(ordaxis[X60]),X60);
  fp.makeorder(&(ordaxis[X120]),X120);
  return 1;
}

mObj2D::update(P2 ss,P2 sp, REAL t) {

  if(!fp.update(ss,sp,t)) return 0;
  fp.makeorder(&(ordaxis[X]),X);
  fp.makeorder(&(ordaxis[X60]),X60);
  fp.makeorder(&(ordaxis[X120]),X120);
  return 1;
}

void 
mObj2D::draw(REAL t) {

  P2 p;
  if(pos(p,t)) {
    
    glPushMatrix();    
    glTranslatef(p[X],p[Y],0.0);
    glRotatef((t-fp.starttime)*thetastep,normvec[X],normvec[Y],0.0);
    gluSphere(ball,radius,20,20);
    glPopMatrix();
    glFinish();
  }
}

REAL
mObj2D::distance2(mObj2D mo2) {

  return dist2(fp.startpoint,mo2.fp.startpoint);
}

REAL
mObj2D::distance2(mObj2D mo2, REAL t) {

  P2 p1,p2;
  if(!pos(p1,t)) return -1;
  if(!mo2.pos(p2,t)) return -1;
  return dist2(p1,p2);
}

REAL
mObj2D::closestcand(mObj2D **mo, REAL t) {

  REAL d;
  REAL dmin = MAX;

  *mo = NULL;
  for(int axis = X; axis<=X120;axis++) {
    
    if(lcand[axis]!=NULL) {
      d = distance2(*lcand[axis],t);
      if(d<dmin) {
	*mo = lcand[axis];
	dmin = d;
      }
    }
  }
  return dmin;
}   
  

REAL
mObj2D::sord(short axis) {

  return ((axis<=X120)?ordaxis[axis].speed:
	               (-ordaxis[axis-3].speed));
}

mObj2D *
mObj2D::inTree1(short axis) {

  bbTree *rt;
  rt = relNode[axis].relTree();
  if(rt==NULL) return NULL;
  return rt->relmo;
}

mObj2D *
mObj2D::inTree2(short axis, short flag) {

  bbTree *rt;
  if(flag==UP)
    rt=cuNode[axis].relTree();
  else 
    rt=clNode[axis].relTree();
  if(rt==NULL) return NULL;
  return rt->relmo;
}

mObj2D *
mObj2D::islcand(short axis) {

  mObj2D *m=inTree1(axis);
  if(m==NULL) return NULL;
  if(m->lcand[axis]==this) return m;
  else return NULL;
}

mObj2D *
mObj2D::inSameCone(mObj2D *mo, short axis) {

  mObj2D *m1 = inTree1(axis);
  mObj2D *m2 = mo->inTree1(axis);
  if((m1!=NULL)&&(m1==m2)) return m1;
  else return NULL;
}

void
mObj2D::resetNodes(void) {

  for(short axis=X;axis<=X120;axis++) {
    
    relNode[axis].makeNode(this,axis);
    cuNode[axis].makeNode(this,axis);
    clNode[axis].makeNode(this,axis);
  }
}


REAL 
mObj2D::hitBoundary(Boundary *b, REAL t, short *side) {

  return fp.hitBoundary(b,t,side);
}

mObj2D::bounced(Boundary *b, REAL t, short side) {

  short a;
  switch(side) {
  case 0:
    a = fp.bounced(b->left,t,X);
    break;
  case 1:
    a = fp.bounced(b->top,t,Y);
    break;
  case 2:
    a = fp.bounced(b->right,t,X);
    break;
  case 3:
    a = fp.bounced(b->bottom,t,Y);
    break;
  }
  if(!a) return 0;
  for(short axis=X;axis<=X120;axis++)
    fp.makeorder(&(ordaxis[axis]),axis);
  return 1;
}


mObj2D::bounced(mObj2D *mo, REAL t) {

  P2 c1,c2,v,v1,v2,ns;
  REAL l;
  if(!pos(c1,t)) return 0;
  if(!mo->pos(c2,t)) return 0;
  REAL d = dist2(c1,c2);

  /*
  REAL sumr = radius+mo->radius;
  sumr *= sumr;
  if(d<sumr-epsilon) return 0;
  if(d>sumr+epsilon) return 0;*/

  v[X] = c2[X]-c1[X];
  v[Y] = c2[Y]-c1[Y];

  l = dotcross(fp.speed,v)/d;
  v1[X] = v[X]*l;
  v1[Y] = v[Y]*l;

  l = dotcross(mo->fp.speed,v)/d;
  v2[X] = v[X]*l;
  v2[Y] = v[Y]*l;
  
  v[X] = v2[X]-v1[X];
  v[Y] = v2[Y]-v1[Y];
  
  ns[X] = fp.speed[X]+v[X];
  ns[Y] = fp.speed[Y]+v[Y];
  fp.update(c1,ns,t);

  ns[X] = mo->fp.speed[X]-v[X];
  ns[Y] = mo->fp.speed[Y]-v[Y];
  mo->fp.update(c2,ns,t);

  for(short axis=X;axis<=X120;axis++) {
    fp.makeorder(&(ordaxis[axis]),axis);
    mo->fp.makeorder(&(mo->ordaxis[axis]),axis);
  }
  return 1;
}


mObj2D::~mObj2D(void) {

}

savedObj* 
mObj2D::save(REAL t) {

  savedObj *a;
  mObj2D *mo;
  a = new savedObj;
  if(!pos(a->currP,t)) {
    
    delete a;
    return NULL;
  }
  a->index = index;
  for(short axis=X;axis<=X120;axis++) {

    mo = inTree1(axis);
    if(mo==NULL) a->inCone[axis] = -1;
    else a->inCone[axis] = mo->index;
    mo = inTree2(axis,UP);
    if(mo==NULL) a->hitUp[axis] = -1;
    else a->hitUp[axis] = mo->index;
    mo = inTree2(axis,DOWN);
    if(mo==NULL) a->hitDown[axis] = -1;
    else a->hitDown[axis] = mo->index;
  }
  return a;
}

REAL
solve(REAL a, REAL b, REAL c, REAL t) {

  REAL x,delta;
  if(a==0) {
    if(b==0) return MAX;
    x = -(c/b);
    if(x>t) return x;
    else return MAX;
  }
  
  delta = b*b-4*a*c;
  if(delta<0) return MAX;
  delta = sqrt(delta);
  if(a>0) {
    x = (-b-delta)/(2*a);
    if(x>t) return x;
    x += delta/a;
    if(x>t) return x;
  } else {
    x = (-b+delta)/(2*a);
    if(x>t) return x;
    x -= delta/a;
    if(x>t) return x;
  }
  return MAX;
}

REAL
solve(REAL a, REAL b, REAL c, REAL t, short nr) {

  REAL x,delta;

  if(nr>=3) return MAX;
  if(a==0) {
    if(nr>=2) return MAX;
    if(b==0) return MAX;
    x = -(c/b);
    if(x>t) return x;
    return MAX;
  }
  
  delta = b*b-4.0*a*c;
  if(delta<0) return MAX;
  delta = sqrt(delta);
  if(delta==0) {

    x = -b/(2*a);
    if((nr<=1)&&(x>t)) return x;
    return MAX;
  }

  if(a>0) {
    x = (-b-delta)/(2*a);
    if((nr==1)&&(x>t)) return x;
    x+=delta/a;
    if(x>t) return x;
  } else {
    x = (-b+delta)/(2*a);
    if((nr==1)&&(x>t)) return x;
    x-=delta/a;
    if(x>t) return x;
  }
  return MAX;
}

REAL
solve(REAL a, REAL b, REAL c, REAL t, short *nnr) {

  return solve(a,b,c,t,1,nnr);
}

    
REAL 
solve(REAL a, REAL b, REAL c, REAL t, short nr, short *nnr) {

  REAL x,delta;

  if(nr>=3) return MAX;
  if(a==0) {
    if(nr>=2) return MAX;
    if(b==0) return MAX;
    x = -(c/b);
    if(x>t) {
      *nnr = 1;
      return x;
    } else return MAX;
  }
  
  delta = b*b-4.0*a*c;
  if(delta<0) return MAX;
  delta = sqrt(delta);
  if(delta==0) {
    
    x = -b/(2*a);
    if((nr<=1)&&(x>t)) {
      *nnr = 1;
      return x;
    }
    return MAX;
  }

  if(a>0) {
    x = (-b-delta)/(2*a);
    if((nr==1)&&(x>t)) {
      *nnr = 1;
      return x;
    } 
    x+=delta/a;
    if(x>t) {
      *nnr = 2;
      return x;
    }
  } else {
    x = (-b+delta)/(2*a);
    if((nr==1)&&(x>t)) {
      *nnr = 1;
      return x;
    }
    x-=delta/a;
    if(x>t) {
      *nnr = 2;
      return x;
    }
  }
  return MAX;
}

inline
void
getPara(mObj2D *mo1, mObj2D *mo2, mObj2D *mo3, mObj2D *mo4,
	REAL *a, REAL *b, REAL *c) {

  P2 *sp1,*ss1,*sp2,*ss2,*sp3,*ss3,*sp4,*ss4;
  P2 msp1, msp2, msp3, msp4;
  REAL dxa1,dxb1,dya1,dyb1,dxa2,dxb2,dya2,dyb2;
  REAL st1,st2,st3,st4;

  sp1 = &((mo1->fp).startpoint);
  ss1 = &((mo1->fp).speed);
  sp2 = &((mo2->fp).startpoint);
  ss2 = &((mo2->fp).speed);
  sp3 = &((mo3->fp).startpoint);
  ss3 = &((mo3->fp).speed);
  sp4 = &((mo4->fp).startpoint);
  ss4 = &((mo4->fp).speed);

  st1 = (mo1->fp).starttime;
  st2 = (mo2->fp).starttime;
  st3 = (mo3->fp).starttime;
  st4 = (mo4->fp).starttime;
  
  msp1[X] = ((*sp1)[X])-st1*((*ss1)[X]);
  msp1[Y] = ((*sp1)[Y])-st1*((*ss1)[Y]);

  msp2[X] = ((*sp2)[X])-st2*((*ss2)[X]);
  msp2[Y] = ((*sp2)[Y])-st2*((*ss2)[Y]);

  msp3[X] = ((*sp3)[X])-st3*((*ss3)[X]);
  msp3[Y] = ((*sp3)[Y])-st3*((*ss3)[Y]);

  msp4[X] = ((*sp4)[X])-st4*((*ss4)[X]);
  msp4[Y] = ((*sp4)[Y])-st4*((*ss4)[Y]);
  
  dxa1 = ((*ss1)[X])-((*ss2)[X]);
  dxb1 = msp1[X]-msp2[X];
  dya1 = ((*ss1)[Y])-((*ss2)[Y]);
  dyb1 = msp1[Y]-msp2[Y];
  dxa2 = ((*ss3)[X])-((*ss4)[X]);
  dxb2 = msp3[X]-msp4[X];
  dya2 = ((*ss3)[Y])-((*ss4)[Y]);
  dyb2 = msp3[Y]-msp4[Y];
  
  *a = (dxa1*dxa1+dya1*dya1-dxa2*dxa2-dya2*dya2);
  *b = 2*(dxa1*dxb1+dya1*dyb1-dxa2*dxb2-dya2*dyb2);
  *c = (dxb1*dxb1+dyb1*dyb1-dxb2*dxb2-dyb2*dyb2);
}

REAL
meet(mObj2D *mo1, mObj2D *mo2, mObj2D *mo3, mObj2D *mo4, REAL t) {

  if((mo1==NULL)||(mo2==NULL)||(mo3==NULL)||(mo4==NULL))
    return MAX;
  REAL a,b,c;
  getPara(mo1,mo2,mo3,mo4,&a,&b,&c);
  return solve(a,b,c,t);
}

REAL
meet(mObj2D *mo1, mObj2D *mo2, mObj2D *mo3, mObj2D *mo4, REAL t, 
     short nr, short *nnr) {
  
  if((mo1==NULL)||(mo2==NULL)||(mo3==NULL)||(mo4==NULL))
    return MAX;
  REAL a,b,c;
  getPara(mo1,mo2,mo3,mo4,&a,&b,&c);
  if(nnr==NULL) return solve(a,b,c,t,nr);
  else return solve(a,b,c,t,nr,nnr);
}

REAL 
bounceTime(mObj2D *mo1, mObj2D *mo2, REAL r, REAL t) {

  REAL sdx,sdy,pdx,pdy;
  P2 *s1,*s2,*p1,*p2;
  REAL s1x,s1y,s2x,s2y,p1x,p1y,p2x,p2y;
  REAL st1, st2;
  REAL a,b,c;
  if(mo1->distance2(*mo2,t)<=r*r) return t; 
  s1 = &(mo1->fp.speed);
  s2 = &(mo2->fp.speed);
  p1 = &(mo1->fp.startpoint);
  p2 = &(mo2->fp.startpoint);
  st1 = mo1->fp.starttime;
  st2 = mo2->fp.starttime;
  s1x = (*s1)[X];
  s1y = (*s1)[Y];
  s2x = (*s2)[X];
  s2y = (*s2)[Y];
  p1x = (*p1)[X];
  p1y = (*p1)[Y];
  p2x = (*p2)[X];
  p2y = (*p2)[Y];
  sdx = s1x - s2x;
  pdx = p1x-p2x-s1x*st1+s2x*st2;
  sdy = s1y - s2y;
  pdy = p1y-p2y-s1y*st1+s2y*st2;
  a = sdx*sdx+sdy*sdy;
  b = 2*(sdx*pdx+sdy*pdy);
  c = pdx*pdx+pdy*pdy-r*r;
  return solve(a,b,c,t);
}


mObjs::mObjs(int size) {

  nofObjs = size;
  if(size>0)
    mos = new mObj2D[size];
  for(int i=0;i<size;i++) mos[i].setindex(i);
  rr = 0;
  clo1 = -1;
  clo2 = -1;
  bT = MAX;
  createBall();
}

mObjs::mObjs(int size, REAL r) {

  nofObjs = size;
  if(size>0)
    mos = new mObj2D[size];
  for(int i=0;i<size;i++) mos[i].setindex(i);
  rr = r;
  clo1 = -1;
  clo2 = -1;
  bT = MAX;
  createBall();
}

mObjs::mObjs(FILE *fin) {
  
  char s[100];
  int i;
  P2 p1,p2;

  if(fgets(s,100,fin)==NULL) return;

  nofObjs = atoi(s);
  if(nofObjs<1) return;

  mos = new mObj2D[nofObjs];
  for(i=0;i<nofObjs;i++) mos[i].setindex(i);
  
  clo1 = -1;
  clo2 = -1;
  fgets(s,100,fin);
  rr = atof(s);
  bT = MAX;

  fgets(s,100,fin);
  sscanf(s,"%f %f %f %f",&b.left,&b.right,&b.bottom,&b.top);

  for(i=0;i<nofObjs;i++) {
    fgets(s,100,fin);
    sscanf(s,"%f %f %f %f",&p1[X],&p1[Y],&p2[X],&p2[Y]);
    mos[i].set(p1,p2,0,rr);
  }
  init();
  createBall();
}  

void
mObjs::setboundary(REAL le, REAL ri, REAL bo, REAL to) {

  b.left = le;
  b.right = ri;
  b.bottom = bo;
  b.top = to;
}

mObjs::~mObjs() {

  delete queue1[0];
  delete queue1[1];
  delete queue1[2];
  delete queue2;
  delete queue3;
  if(nofObjs>0)
    delete [] mos;
  gluDeleteQuadric(ball);

}

void
mObjs::reset(int size) {

  if(nofObjs == 0) {

    nofObjs = size;
    mos = new mObj2D[size];
  } else {

    delete [] mos;
    nofObjs = size;
    if(size>0) 
      mos = new mObj2D[size];
  }
  for(int i=0;i<size;i++) mos[i].setindex(i);
}

void
mObjs::generate(REAL ls, REAL us) {

  int i;
  REAL left = b.left+rr;
  REAL bottom = b.bottom+rr;
  REAL length = b.right-b.left-2*rr;
  REAL height = b.top-b.bottom-2*rr;
  REAL is = us-ls;
  P2 p1,p2;
  for(i=0;i<nofObjs;i++) {

    p1[X] = (rand()*1.0/32767)*length+left;
    p1[Y] = (rand()*1.0/32767)*height+bottom;
    p2[X] = (rand()*1.0/32767)*is+ls;
    p2[Y] = (rand()*1.0/32767)*is+ls;
    mos[i].set(p1,p2,0,rr);
  }
  init();
}


int
mObjs::gen_nocollision(REAL ls, REAL us, int stopRounds) {

  int i;
  REAL left = b.left+rr;
  REAL bottom = b.bottom+rr;
  REAL length = b.right-b.left-2*rr;
  REAL height = b.top-b.bottom-2*rr;
  REAL is = us-ls;
  int failureround=0;
  P2 p1,p2;
  for(i=0;i<nofObjs;i++) {

    do {
      p1[X] = (rand()*1.0/32767)*length+left;
      p1[Y] = (rand()*1.0/32767)*height+bottom;
      failureround++;
    } while (collide(p1,i)&&(failureround<stopRounds));
    if(failureround==stopRounds) return 0;
    failureround=0;
    if(rand()&1)
      p2[X] = (rand()*1.0/32767)*is+ls;
    else
      p2[X] = -(rand()*1.0/32767)*is-ls;
    if(rand()&1)
      p2[Y] = (rand()*1.0/32767)*is+ls;
    else
      p2[Y] = -(rand()*1.0/32767)*is-ls;
    mos[i].set(p1,p2,0,rr);
  } 
  init();
  return 1;
}

int
mObjs::collide(P2 p, int number) {

  P2 tmpp;
  REAL squarer = 4*rr*rr;
  if(number==0) return 0;
  if(number>=nofObjs) return -1;
  for(int i=0;i<number;i++) {
    tmpp[X] = mos[i].fp.startpoint[X];
    tmpp[Y] = mos[i].fp.startpoint[Y];
    if(dist2(tmpp,p)<=squarer) return 1;
  }
  return 0;
}

void
mObjs::init(void) {

  int i;
  short axis;

  for(axis=X; axis<=X120; axis++) {    
    queue1[axis] = new eventQueue1(this,axis);
  }

  for(i=0;i<nofObjs;i++) 
       mos[i].resetNodes();

  for(axis=X; axis<=X120; axis++) {
    init(axis);
    setlcand(axis);
  }
  queue2 = new eventQueue2(this);
  queue3 = new eventQueue3(this);

}

void 
mObjs::init(short axis) {

  int i;

  bbTree btCurr((short)axis),btH((short)axis),
    btL((short) axis),btHC((short)axis);
  
  mObj2D *currP;
  bbTNode *lEdge,*uEdge;

  for(i=0;i<nofObjs;i++) {

    currP = queue1[axis]->lnode[i]->org;
    btCurr.split(&btL,&btHC,currP,axis+1);
    btHC.split(&(currP->cand[axis]),&btH,currP,axis+2);
    lEdge = btL.max();
    if(lEdge!=NULL)
      lEdge->point->parentU[axis].insert(&(currP->cuNode[axis]),axis);
    uEdge = btH.min();
    if(uEdge!=NULL)
      uEdge->point->parentL[axis].insert(&(currP->clNode[axis]),axis);
    btCurr.join(&btL,&btH,&(currP->relNode[axis]));
  }
  btCurr.decompose();
}


void
mObjs::setlcand(short axis) {

  int i;
  for(i=0;i<nofObjs;i++) {

      if(mos[i].cand[axis].root!=NULL) 
	mos[i].lcand[axis] = mos[i].cand[axis].root->mindes->point;
      else
	mos[i].lcand[axis] = NULL;
  }
}

void
mObjs::setlcand(void) {
  
  int axis;
  for(axis=X;axis<=X120;axis++) setlcand(axis);
}

REAL
mObjs::bclosest(mObj2D **obj1, mObj2D **obj2, REAL t) {

  int i,j;
  REAL tempd, d=MAX;
  if(nofObjs==0) return 0;
  *obj1 = NULL;
  *obj2 = NULL;
  for(i=0;i<nofObjs;i++) {
    for(j=0;j<i;j++) {
      tempd = mos[i].distance2(mos[j],t);
      if(tempd<d) {
	d = tempd;
	*obj1 = &(mos[i]);
	*obj2 = &(mos[j]);
      }
    }
  }
  return sqrt(d);
}

REAL
mObjs::processevent(mObj2D** obj1, mObj2D** obj2, REAL *t, REAL *nt) {

  REAL T;
  short eventtype, axis;
  mObj2D *mo1, *mo2;
  P2 p1, p2;

  T = nextcritical(&mo1,&mo2,&eventtype,&axis);
  *t = T;
  if(T == MAX) return -1;
  if(eventtype == EVE1) {
    
    queue1[axis]->process(axis,T);
    exchange(mo1,mo2,axis,T);
    shed(mo2,mo1,axis,T);
    mo1->pos(p1,T);
    mo2->pos(p2,T);
    if(orderY(p1,axis)<=orderY(p2,axis)) {
      comeOutB(mo2,mo1,axis+2,T);
      comeInU(mo2,mo1,axis+1,T);
    } else { 
      comeOutU(mo1,mo2,axis+1,T);
      comeInB(mo1,mo2,axis+2,T);
    }
  }
  else if(eventtype == EVE2)  
    queue2->process(this);
  else if(eventtype == EVE3)
    bounced(mo1,T,axis);
  else if(eventtype == EVE4)
    bounced(mo1,mo2,T);
  int clind1, clind2;
  queue2->extractMinPair(this,obj1,obj2);
  clind1 = (*obj1)->index;
  clind2 = (*obj2)->index;
  if(eventtype == EVE4)
    bT = MAX;
  else {
    if(((eventtype==EVE3)&&((mo1->index == clind1)||(mo1->index == clind2)))
       || !samepair(clo1,clo2,clind1,clind2)) {
      clo1 = clind1;
      clo2 = clind2;
      bT = bounceTime(*obj1,*obj2,2*rr,T);
    } 
  }
  *nt = nextcritical();
  return sqrt((*obj1)->distance2(**obj2, T));
}

REAL
mObjs::getcpair(mObj2D **obj1, mObj2D **obj2,short eventtype,REAL t) {
  
  int clind1, clind2;
  queue2->extractMinPair(this,obj1,obj2);
  clind1 = (*obj1)->index;
  clind2 = (*obj2)->index;
  if(eventtype == EVE4)
    bT = MAX;
  else {
    if(!samepair(clo1,clo2,clind1,clind2) ||
       eventtype == EVE3) {
      clo1 = clind1;
      clo2 = clind2;
      bT = bounceTime(*obj1,*obj2,2*rr,t);
    } 
  }
  return sqrt((*obj1)->distance2(**obj2, t));
}

REAL
mObjs::nextcritical(void){

  int i;
  REAL T = MAX, tempT;

  for(i=X;i<=X120;i++) {
    tempT = queue1[i]->extractMin();
    if(tempT<T) 
      T = tempT;
  }

  tempT = queue2->extractMin();
  if(tempT<T) 
    T = tempT;
  
  tempT = queue3->extractMin();
  if(tempT<T) T = tempT;

  if(bT<T) T=bT;

  return T;
}

REAL
mObjs::nextcritical(mObj2D **mo1, mObj2D **mo2, short *type, short *axis) {

  int eventtype,ax,i;
  REAL T = MAX, tempT;

  eventtype = EVE1;

  for(i=X;i<=X120;i++) {
    tempT = queue1[i]->extractMin();
    if(tempT<T) {
      ax = i;
      T = tempT;
    }
  }

  tempT = queue2->extractMin();
  if(tempT<T) {
    T = tempT;
    eventtype = EVE2;
  }

  tempT = queue3->extractMin();
  if(tempT<T) {
    T = tempT;
    eventtype = EVE3;
  }
  
  if(bT<T) {
    T = bT;
    eventtype = EVE4;
  }

  if(T==MAX) return MAX;
  *type = eventtype;
  *axis = ax;

  if(eventtype == EVE1)    
    queue1[ax]->extractMin(mo1, mo2);
  else if(eventtype == EVE2) 
    queue2->extractMin(this,mo1,mo2,axis);
  else if(eventtype == EVE3)
    queue3->extractMin(this,mo1,axis);
  else if(eventtype == EVE4) {
    *mo1 = &(mos[clo1]);
    *mo2 = &(mos[clo2]);
  }
  return T;  
}

void
mObjs::update(mObj2D *mo, mObj2D *ncand, short axis, REAL t) {

  if(ncand == mo->lcand[axis]) return ;
  mo->lcand[axis] = ncand;
  queue2->update(this, mo->index, axis, t);
}

void
mObjs::update(mObj2D *mo, short axis, REAL t) {

  mObj2D *ncand;
  bbTNode *min;
  min = mo->cand[axis].minSkey();
  if(min==NULL) ncand=NULL;
  else ncand = min->point;
  update(mo,ncand,axis,t);
}

void
mObjs::exchange(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  bbTNode *mr1, *mr2, *curr;
  mObj2D *mopa = mo1->inSameCone(mo2,axis);
  int key;
  if(mopa==NULL) return;
  mr1 = &(mo1->relNode[axis]);
  mr2 = &(mo2->relNode[axis]);
  curr = mr1;

  key = mo1->ord(axis);

  while(curr!=NULL) {
    if(curr->mindes!=mr1) break;
    curr->minskey = key;
    curr = curr->parent;
  }

  while(curr!=NULL) {
    if(curr->mindes!=mr2) break;
    curr->minskey = key;
    curr->mindes = mr1;
    curr= curr->parent;
  }
  if(curr==NULL) update(mopa,axis,t);
}

void
mObjs::comeInU(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  if(axis<=X120) comeInUp(mo1,mo2,axis,t);
  else comeInUp(mo2,mo1,axis-3,t);
}

void 
mObjs::comeInB(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  if(axis<=X120) comeInBelow(mo1,mo2,axis,t);
  else comeInBelow(mo2,mo1,axis-3,t);
}

void 
mObjs::comeOutU(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  if(axis<=X120) comeOutUp(mo1,mo2,axis,t);
  else comeOutUp(mo2,mo1,axis-3,t);
}

void 
mObjs::comeOutB(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  if(axis<=X120) comeOutBelow(mo1,mo2,axis,t);
  else comeOutBelow(mo2,mo1,axis-3,t);
}

  
void
mObjs::comeInUp(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  bbTree btP((short)axis);
  bbTNode *edge;

  // cout<<"\n Enter ComeInUp"<<mo1->index<<"  "<<mo2->index<<"  "<<axis<<"  ";
  /* judge whether need modify data structure */
  if(mo2->inTree2(axis,DOWN)!=mo1) return;

  /* remove mo1 from old cone */
  mObj2D *oldCone = mo1->inTree1(axis);
  if(oldCone!=NULL) {
    oldCone->cand[axis].del(&(mo1->relNode[axis]));
    update(oldCone,axis,t);
  }

  /* insert mo1 into mo2's cone */

  mo1->relNode[axis].reset(axis);
  mo2->cand[axis].insert(&(mo1->relNode[axis]),axis+1);
  update(mo2,axis,t);

  /* delete mo2 from mo1's parents */
  mo1->parentL[axis].del(&(mo2->clNode[axis]));
  mo2->clNode[axis].reset(axis);

  /* update mo2's parent */
  mo1->parentU[axis].split(&btP,mo2,axis,1);
  mo2->parentU[axis].ljoinwith(&btP);

  /* find the edge mo2 hits */
  edge = mo1->parentU[axis].min();
  mObj2D *p;
  if(edge!=NULL) p = edge->point;
  else p = mo1->inTree2(axis,DOWN);
  if(p!=NULL) p->parentL[axis].insert(&(mo2->clNode[axis]),axis);
  //  cout<<"\n Leave comeInUP \n";
}

void
mObjs::comeInBelow(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  bbTree btP((short)axis);
  bbTNode *edge;

  //  cout<<"\n Enter ComeInBelow"<<mo1->index<<"  "<<mo2->index<<"  "<<axis<<"  ";
 
  /* judge whether need modify data structure */
  if(mo2->inTree2(axis,UP)!=mo1) return;

  /* remove mo1 from old cone */
  mObj2D *oldCone = mo1->inTree1(axis);
  if(oldCone!=NULL) {
    oldCone->cand[axis].del(&(mo1->relNode[axis]));
    update(oldCone,axis,t);
  }

  /* insert mo1 into mo2's cone */

  mo1->relNode[axis].reset(axis);
  mo2->cand[axis].insert(&(mo1->relNode[axis]),axis+1);
  update(mo2,axis,t);

  /* delete mo2 from mo1's parents */
  mo1->parentU[axis].del(&(mo2->cuNode[axis]));
  mo2->cuNode[axis].reset(axis);

  /* update mo2's parent */
  mo1->parentL[axis].split(&btP,mo2,axis,1);
  mo2->parentL[axis].ljoinwith(&btP);

  /* find the edge mo2 hits */
  edge = mo1->parentL[axis].min();
  mObj2D *p;

  if(edge!=NULL) p = edge->point;
  else p = mo1->inTree2(axis,UP);
  if(p!=NULL) 
    p->parentU[axis].insert(&(mo2->cuNode[axis]),axis);
  //  cout<<"\n Leave comeInBelow \n";
}

void
mObjs::comeOutUp(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  bbTree btP((short)axis);
  bbTNode *edge;
  mObj2D *newcone; 

  //  cout<<"\n Enter ComeOutUp"<<mo1->index<<"  "<<mo2->index<<"  "<<axis<<"  ";
  /* find whether mo1 is in mo2's cone */
  if(mo1->inTree1(axis)!=mo2) return;

  mo2->parentU[axis].split(&btP,mo1,axis+1,0);
  mo1->parentU[axis].rjoinwith(&btP);

  mo2->cand[axis].del(&(mo1->relNode[axis]));
  mo1->relNode[axis].reset(axis);
  update(mo2,axis,t);

  edge = mo2->parentU[axis].max();
  if(edge==NULL) newcone = mo2->inTree1(axis);
  else newcone = edge->point;

  if(newcone!=NULL) {
     newcone->cand[axis].insert(&(mo1->relNode[axis]),axis+1);
     update(newcone,axis,t);
  }
 
  newcone = mo2->inTree2(axis,DOWN);
  if(newcone!=NULL) newcone->parentL[axis].del(&(mo2->clNode[axis]));
  mo2->clNode[axis].reset(axis);
  mo1->parentL[axis].insert(&(mo2->clNode[axis]),axis);
  //  cout<<"\n Leave comeOutUP \n";
}

void
mObjs::comeOutBelow(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  bbTree btP((short)axis);
  bbTNode *edge;
  mObj2D *newcone; 

  //  cout<<"\n Enter ComeOutBelow"<<mo1->index<<"  "<<mo2->index<<"  "<<axis<<"  ";
  /* find whether mo1 is in mo2's cone */
  if(mo1->inTree1(axis)!=mo2) return;

  mo2->parentL[axis].split(&btP,mo1,axis+5,0);
  mo1->parentL[axis].rjoinwith(&btP);

  mo2->cand[axis].del(&(mo1->relNode[axis]));
  mo1->relNode[axis].reset(axis);
  update(mo2,axis,t);

  edge = mo2->parentL[axis].max();
  if(edge==NULL) newcone = mo2->inTree1(axis);
  else newcone = edge->point;

  if(newcone!=NULL) {
     newcone->cand[axis].insert(&(mo1->relNode[axis]),axis+1);
     update(newcone,axis,t);
  }
 
  newcone = mo2->inTree2(axis,UP);
  if(newcone!=NULL) newcone->parentU[axis].del(&(mo2->cuNode[axis]));
  mo2->cuNode[axis].reset(axis);
  mo1->parentU[axis].insert(&(mo2->cuNode[axis]),axis);
  //  cout<<"\n Leave comeOutBelow \n";
}

void
mObjs::shed(mObj2D *mo1, mObj2D *mo2, short axis, REAL t) {

  bbTree btTemp((short)axis);
  bbTNode *edge;
  mObj2D *mpa;

  //  cout<<"\n Enter Shed"<<mo1->index<<"  "<<mo2->index<<"  "<<axis<<"  ";  
  if(mo1->inTree2(axis,UP) == mo2) {

    /* split cone of mo2 */
    mo2->cand[axis].split(&btTemp,mo1,axis+1,0);

    /* join appropriate part with cone of mo1 */
    mo1->cand[axis].rjoinwith(&btTemp);    

    /* update parent information */

    /* delete mo1 from mo2's parent */
    mo2->parentU[axis].del(&(mo1->cuNode[axis]));
    mo1->cuNode[axis].reset(axis);

    /* hit mo2 to edge of mo1 */
    mpa = mo2->inTree2(axis,DOWN);
    if(mpa!=NULL) mpa->parentL[axis].del(&(mo2->clNode[axis])); 
    mo2->clNode[axis].reset(axis);
    mo1->parentL[axis].insert(&(mo2->clNode[axis]),axis);
 
    /* find the edge mo1 hit  */
    edge = mo2->cand[axis].max();
    mObj2D *np;
    if(edge!=NULL) np = edge->point;
    else  np = mo2->inTree2(axis,UP);
    if(np!=NULL)
      np->parentU[axis].insert(&(mo1->cuNode[axis]),axis);

  } else if(mo1->inTree2(axis,DOWN) == mo2) {

   /* split cone of mo2 */
    mo2->cand[axis].split(&btTemp,mo1,axis+2,1);

    /* join appropriate part with cone of mo1 */
    mo1->cand[axis].ljoinwith(&btTemp);    

    /* update parent information */

    /* delete mo1 from mo2's parent */
    mo2->parentL[axis].del(&(mo1->clNode[axis]));
    mo1->clNode[axis].reset(axis);

    /* hit mo2 to edge of mo1 */
    mpa = mo2->inTree2(axis,UP);
    if(mpa!=NULL) mpa->parentU[axis].del(&(mo2->cuNode[axis]));
    mo2->cuNode[axis].reset(axis);
    mo1->parentU[axis].insert(&(mo2->cuNode[axis]),axis);
 
    /* find the edge mo1 hit  */
    edge = mo2->cand[axis].min();
    mObj2D *np;
    if(edge!=NULL) np = edge->point;
    else  np = mo2->inTree2(axis,DOWN);
    if(np!=NULL)
      np->parentL[axis].insert(&(mo1->clNode[axis]),axis);

  } else return;
  
  update(mo1,axis,t);
  update(mo2,axis,t);
  //  cout<<"\n Leave Shed \n";
}

void
mObjs::reschedule(mObj2D *mo, REAL t) {

  short axis;
  mObj2D *mp;
  int ind_q1;
  int ind = mo->index;
  for(axis = X; axis<=X120; axis++) {
    ind_q1 = mo->ind_q1[axis];
    queue1[axis]->reschedule(ind_q1-1,t);
    queue1[axis]->reschedule(ind_q1,t);
  }
   
  queue2->update(this,ind,X,t);
  queue2->update(this,ind,X120,t);
  
  for(axis = X; axis<= X120; axis++) {
    mp = mo->islcand(axis);
    if(mp!=NULL) 
      queue2->update(this,mp->index,axis,t);
  }

  queue3->update(this,ind,t);
}

mObjs::bounced(mObj2D* mo, REAL t, short side) {

  short a;

  a = mo->bounced(&b,t,side);
  if(!a) return 0;
  reschedule(mo,t);
  return 1;
}

mObjs::bounced(mObj2D *mo1, mObj2D *mo2, REAL t) {

  if(!mo1->bounced(mo2,t)) return 0;
  reschedule(mo1,t);
  reschedule(mo2,t);
  return 1;
}

void
mObjs::drawCones(REAL t, short axis) {

  int i;
  glBegin(GL_LINES);
  for(i=0;i<nofObjs;i++) {

    drawUpEdge(i,t,axis);
    drawBelowEdge(i,t,axis);
  }
  glColor3f(1.0,0.0,0.0);
  for(i=0;i<nofObjs;i++) 
    drawParent(i,t,axis);
  glEnd();
}

void
mObjs::drawUpEdge(int index, REAL t, short axis) {

  P2 p;
  P2 p0, p1;
  mObj2D *mp;
  mp = mos[index].inTree2(axis,DOWN);
  if(!mos[index].pos(p0,t)) return;
  if(mp==NULL) {

    switch(axis) {
      
    case X:
      p[X] = b.right;
      p[Y] = S_3_3*(b.right-p0[X])+p0[Y];
      break;
    case X60:
      p[X] = p0[X];
      p[Y] = b.top;
      break;
    case X120:
      p[X] = b.left;
      p[Y] = -S_3_3*(b.left-p0[X])+p0[Y];
      break;
    }
  } else {
    
    if(!mp->pos(p1,t)) return;
    intersect(p0,p1,p,axis);
  }
  prune(p0,p,b);
  glVertex2fv(p0);
  glVertex2fv(p);
}

void
mObjs::drawBelowEdge(int index, REAL t, short axis) {

  P2 p;
  P2 p0, p1;
  mObj2D *mp;
  mp = mos[index].inTree2(axis,UP);
  if(!mos[index].pos(p0,t)) return;

  if(mp==NULL) {

    switch(axis) {
      
    case X:
      p[X] = b.right;
      p[Y] = -S_3_3*(b.right-p0[X])+p0[Y];
      break;
    case X60:
      p[X] = b.right;
      p[Y] = S_3_3*(b.right-p0[X])+p0[Y];
      break;
    case X120:
      p[X] = p0[X];
      p[Y] = b.top;
      break;
    }      
  } else {
    
    if(!mp->pos(p1,t)) return;
    intersect(p1,p0,p,axis);
  }
  prune(p0,p,b);
  glVertex2fv(p0);
  glVertex2fv(p);
}

void
mObjs::drawParent(int index, REAL t, short axis) {

  mObj2D *mp = mos[index].islcand(axis);
  if(mp==NULL) return;
  P2 p0, p;
  if(!mos[index].pos(p0,t)) return;
  if(!mp->pos(p,t)) return;
  glVertex2fv(p0);
  glVertex2fv(p);
}

void
mObjs::save(FILE* fout) {
  
  fprintf(fout,"%d\n",nofObjs);
  fprintf(fout,"%f\n",rr);
  fprintf(fout,"%f %f %f %f\n",b.left,b.right,b.bottom,b.top);
  for(int i=0;i<nofObjs;i++) 
    fprintf(fout,"%f %f %f %f\n",
	    mos[i].fp.startpoint[X], mos[i].fp.startpoint[Y], 
	    mos[i].fp.speed[X], mos[i].fp.speed[Y]);
}

savedObjs*
mObjs::save(REAL t) {

  savedObjs *a;
  mObj2D *mo1, *mo2;
  a = new savedObjs;
  a->nofObjs = nofObjs;
  a->mos = (savedObj **) malloc( nofObjs*sizeof(savedObj *));
  for(int i=0;i<nofObjs;i++)
    a->mos[i] = mos[i].save(t);
  nextcritical(&mo1,&mo2,&(a->et),&(a->ax));
  if(mo1==NULL) a->cind1 = -1;
  else a->cind1 = mo1->index;
  if(a->et==EVE3) a->cind2 = -1;
  else {
    if(mo2==NULL) a->cind2 = -1;
    else a->cind2 = mo2->index;
  }
  return a;
}

void
mObjs::drawSaved(savedObj *so) {

  /* draw the ball */
  glPushMatrix();
  glTranslatef(so->currP[X],so->currP[Y],0.0);
  glPopMatrix();
}

void
mObjs::drawSaved(savedObjs *sos, savedObj *so) {

  glColor3f(0.0,0.0,1.0);
  drawSaved(so);

  glColor3f(0.0,0.0,0.0);
  /* draw the cone */

  P2 p0,p1,p;
  int ind;

  glBegin(GL_LINES);
  ind = so->hitDown[X];
  p0[X] = so->currP[X];
  p0[Y] = so->currP[Y];
  if(ind==-1) {

    if(b.right<=S_3*(b.top-p0[Y])+p0[X]) {
      
      p[X] = b.right;
      p[Y] = S_3_3*(b.right-p0[X])+p0[Y];
    } else {
      p[X] = S_3*(b.top-p0[Y])+p0[X];
      p[Y] = b.top;
    } 
  } else {
    
    p1[X] = sos->mos[ind]->currP[X];
    p1[Y] = sos->mos[ind]->currP[Y];
    intersect(p0,p1,p,X);
  }
  glVertex2fv(p0);
  glVertex2fv(p);

  ind = so->hitUp[X];
  if(ind==-1) {

    if(b.right<=-S_3*(b.bottom-p0[Y])+p0[X]) {
      
      p[X] = b.right;
      p[Y] = -S_3_3*(b.right-p0[X])+p0[Y];
    } else {
      p[X] = -S_3*(b.bottom-p0[Y])+p0[X];
      p[Y] = b.bottom;
    } 
  } else {

    p1[X] = sos->mos[ind]->currP[X];
    p1[Y] = sos->mos[ind]->currP[Y];
    intersect(p1,p0,p,X);
  }

  glVertex2fv(p0);
  glVertex2fv(p);
  glEnd();

  ind = so->inCone[X];
  if(ind==-1) return;
  glColor3f(1.0,0.0,0.0);
  glBegin(GL_LINES);
  p[X] = sos->mos[ind]->currP[X];
  p[Y] = sos->mos[ind]->currP[Y];
  glVertex2fv(p0);
  glVertex2fv(p);
  glEnd();
}

void 
mObjs::drawSaved(savedObjs* sos) {
 
  for(int i=0;i<nofObjs;i++) 
    drawSaved(sos,sos->mos[i]);
  glColor3f(1.0,0.0,0.0);
  if(sos->cind1!=-1) drawSaved(sos->mos[sos->cind1]);
  if(sos->et!=EVE3) {
    glColor3f(0.0,0.0,0.0);
    if(sos->cind2!=NULL) drawSaved(sos->mos[sos->cind2]);
  }
}

void 
buildup(innerNode *ia, int lsize) {
  
  int i,l,r;
  for(i=lsize-2;i>=0;i--) {

    l = LS(i); r = RS(i);
    if(ia[l].T<=ia[r].T) {
      ia[i].T = ia[l].T;
      ia[i].ind = ia[l].ind;
    } else {
      ia[i].T = ia[r].T;
      ia[i].ind = ia[r].ind;
    }    
  }
}

void
updateup(innerNode *ia, int index, int lsize) {

  int ne,pa;
  if(index<0) return;
  if(index>=lsize) return;
  int i=index+lsize-1;
  while(i!=0) {
    ne = NE(i);
    pa = PA(i);
    if(ia[i].T<=ia[ne].T) {
      ia[pa].T = ia[i].T;
      ia[pa].ind = ia[i].ind;
    } else {
      ia[pa].T = ia[ne].T;
      ia[pa].ind = ia[ne].ind;
    }
    i = PA(i);
  }
}

eventQueue1::eventQueue1(int size) {

  lsize = size;
  qsize = 2*lsize-3;
  if(qsize>0)
    inode = (innerNode *) malloc(qsize*sizeof(innerNode));
  if(lsize>0)
    lnode = (mObj1D **) malloc(lsize*sizeof(mObj1D*));
}

eventQueue1::eventQueue1(mObjs *mos, short axis) {

  lsize = 0;
  qsize = 0;
  set(mos,axis);
}

eventQueue1::~eventQueue1(void) {

  if(lsize>0) delete lnode;
  if(qsize>0) delete inode;
}

eventQueue1::set(mObjs *mos,short axis) {

  int nos = mos->nofObjs;
  if(nos<=2) return 0;
  if(lsize!=nos) {
    if(lsize>0) delete lnode;
    if(qsize>0) delete inode;
    lsize = nos;
    qsize = 2*lsize - 3;
    inode = (innerNode *) malloc(qsize*sizeof(innerNode));
    lnode = (mObj1D **) malloc(lsize*sizeof(mObj1D *));
  }

  REAL *olist = (REAL*) malloc(nos*sizeof(REAL));
  REAL *olist2 = (REAL*) malloc(nos*sizeof(REAL));
  int *slist = (int*) malloc(nos*sizeof(int));
  mObj1D **molist = (mObj1D**) malloc(nos*sizeof(mObj1D*));
  int i,j;

  for(i=0;i<nos;i++) {
    olist[i] = mos->mos[i].ordaxis[axis].startpoint;
    olist2[i] = mos->mos[i].ordaxis[axis].speed;
    molist[i] = &(mos->mos[i].ordaxis[axis]);
  }
  
  sort2(olist,olist2,slist,nos,DECREASING);
  
  for(i=0;i<lsize;i++) {
    lnode[i] = molist[slist[i]];
    lnode[i]->org->ind_q1[axis] = i;
  }

  for(i=0,j=lsize-2;i<lsize-1;i++,j++) {
    
    inode[j].T = lnode[i]->meet(*lnode[i+1],0);
    inode[j].ind = i;
  }
  
  buildup(inode,lsize-1);

  delete olist;
  delete olist2;
  delete slist;
  delete molist;
  return 1;
}

eventQueue1::set(mObjs *mos,short axis,REAL t) {

  int nos = mos->nofObjs;
  if(nos<=2) return 0;
  if(lsize!=nos) {
    if(lsize>0) delete lnode;
    if(qsize>0) delete inode;
    lsize = nos;
    qsize = 2*lsize - 3;
    inode = (innerNode *) malloc(qsize*sizeof(innerNode));
    lnode = (mObj1D **) malloc(lsize*sizeof(mObj1D *));
  }
  
  REAL *olist = (REAL*) malloc(nos*sizeof(REAL));
  REAL *olist2 = (REAL*) malloc(nos*sizeof(REAL));
  int *slist = (int*) malloc(nos*sizeof(int));
  mObj1D **molist = (mObj1D**) malloc(nos*sizeof(mObj1D*));
  int i,j;

  for(i=0;i<nos;i++) {
    olist[i] = mos->mos[i].ordaxis[axis].pos(t);
    olist2[i] = mos->mos[i].ordaxis[axis].speed;
    molist[i] = &(mos->mos[i].ordaxis[axis]);
  }
  
  sort2(olist,olist2,slist,nos,DECREASING);
  
  for(i=0;i<lsize;i++) {
    lnode[i] = molist[slist[i]];
    lnode[i]->org->ind_q1[axis] = i;
  }

  for(i=0,j=lsize-2;i<lsize-1;i++,j++) {
    
    inode[j].T = lnode[i]->meet(*lnode[i+1],t);
    inode[j].ind = i;
  }
  
  buildup(inode,lsize-1);

  delete olist;
  delete olist2;
  delete slist;
  delete molist;
  return 1;
}

REAL
eventQueue1::extractMin() {
  
  if(qsize<=0) return MAX;
  return inode[0].T;
}

REAL
eventQueue1::extractMin(mObj2D **mo1, mObj2D **mo2) {

  if(qsize<=0) return MAX;
  if(inode[0].T!=MAX) {
    int ind = inode[0].ind;
    *mo1 = lnode[ind]->org;
    *mo2 = lnode[ind+1]->org;
  }      
  return inode[0].T;
}
			  
void
eventQueue1::update(int index) {

  updateup(inode,index,lsize-1);
}

void
eventQueue1::reschedule(int index, REAL T) {

  if((index<0)||(index>=lsize-1)) return;
  inode[index+lsize-2].T = lnode[index]->meet(*lnode[index+1],T);
  update(index);
} 

short
eventQueue1::process() {

  REAL T = inode[0].T;
  int i = inode[0].ind;
  mObj1D *temp;

  if(T==-1) return 0;
  if(T == MAX) return 0;

  temp = lnode[i];
  lnode[i] = lnode[i+1];
  lnode[i+1] = temp;

  reschedule(i-1,T);
  reschedule(i,T);
  reschedule(i+1,T);
  return 1;
}

short
eventQueue1::process(short axis) {

  REAL T = inode[0].T;
  int i = inode[0].ind;
  mObj1D *temp;

  if(T==-1) return 0;
  if(T == MAX) return 0;

  temp = lnode[i];
  lnode[i] = lnode[i+1];
  lnode[i+1] = temp;
  
  lnode[i]->org->ind_q1[axis] = i;
  lnode[i+1]->org->ind_q1[axis] = i+1;
  lnode[i]->org->relNode[axis].updateup(axis);
  lnode[i+1]->org->relNode[axis].updateup(axis);  
  reschedule(i-1,T);
  reschedule(i,T);
  reschedule(i+1,T);
  return 1;
}

short
eventQueue1::process(short axis, REAL t) {

  int i = inode[0].ind;

  mObj1D *temp;

  //  if(lnode[i]->speed>=lnode[i+1]->speed) return 0;

  temp = lnode[i];
  lnode[i] = lnode[i+1];
  lnode[i+1] = temp;
  
  lnode[i]->org->ind_q1[axis] = i;
  lnode[i+1]->org->ind_q1[axis] = i+1;
  lnode[i]->org->relNode[axis].updateup(axis);
  lnode[i+1]->org->relNode[axis].updateup(axis);  

  reschedule(i-1,t);
  reschedule(i,t);
  reschedule(i+1,t);
  return 1;
}

eventQueue2::eventQueue2(mObjs *mos) {

  sizeTi = 0;
  sizeTl = 0;
  sizeQ = 0;
  init(mos);
}

void
eventQueue2::init(mObjs *mos) {

  int nos = mos->nofObjs;    
  if(nos<=2) return;
  nos *= 3;
  if(nos!=sizeTl) {
    if(sizeTi>0) delete inodeT;
    if(sizeTl>0) delete lnodeT;
    if(sizeQ>0) delete inodeQ;
    sizeTl = nos;
    sizeTi = 2*nos-1;
    sizeQ = 2*nos-3;
    inodeT = (innerNode *) malloc(sizeTi*sizeof(innerNode));
    lnodeT = (mObj2D **) malloc(sizeTl*sizeof(mObj2D *));
    inodeQ = (innerNode *) malloc(sizeQ*sizeof(innerNode));
  }
  int i,j,ind1,ind2;

  for(i=0,j=sizeTl-1;i<nos;i++,j++) {
    
    lnodeT[i] = mos->mos[i/3].lcand[i%3];
    inodeT[j].ind = i;
    if(lnodeT[i]==NULL)
      inodeT[j].T = MAX;
    else
      inodeT[j].T = mos->mos[i/3].distance2(*lnodeT[i],0);
  }


  buildup(inodeT,sizeTl);

  for(i=1,j=sizeTl-2;i<sizeTi-1;i+=2,j++) {

    ind1 = inodeT[i].ind;
    ind2 = inodeT[i+1].ind;
    inodeQ[j].T = meet(&(mos->mos[ind1/3]),lnodeT[ind1],
		       &(mos->mos[ind2/3]),lnodeT[ind2],0,1,&(inodeQ[j].un));
    inodeQ[j].ind = i;
  }
  buildup(inodeQ,sizeTl-1);
}

void
eventQueue2::init(mObjs *mos, REAL t) {

  int nos = mos->nofObjs;    
  if(nos<=2) return;    
  nos *= 3;
  if(nos!=sizeTl) {
    if(sizeTi>0) delete inodeT;
    if(sizeTl>0) delete lnodeT;
    if(sizeQ>0) delete inodeQ;
    sizeTl = nos;
    sizeTi = 2*nos-1;
    sizeQ = 2*nos-3;
    inodeT = (innerNode *) malloc(sizeTi*sizeof(innerNode));
    lnodeT = (mObj2D **) malloc(sizeTl*sizeof(mObj2D *));
    inodeQ = (innerNode *) malloc(sizeQ*sizeof(innerNode));
  }
  int i,j,ind1,ind2;

  for(i=0,j=sizeTl-1;i<nos;i++,j++) {

    lnodeT[i] = mos->mos[i/3].lcand[i%3];
    inodeT[j].ind = i;
    if(lnodeT[i]==NULL)
      inodeT[j].T = MAX;
    else
      inodeT[j].T = mos->mos[i/3].distance2(*lnodeT[i],t);
  }

  buildup(inodeT,sizeTl);

  for(i=1,j=sizeTl-2;i<sizeTi-1;i+=2,j++) {

    ind1 = inodeT[i].ind;
    ind2 = inodeT[i+1].ind;
    inodeQ[j].T = meet(&(mos->mos[ind1/3]),lnodeT[ind1],
		       &(mos->mos[ind2/3]),lnodeT[ind2],t,1,&(inodeQ[j].un));
    inodeQ[j].ind = i;
  }
  buildup(inodeQ,sizeTl-1);
}
  
eventQueue2::~eventQueue2(void) {

  if(sizeTl>0) delete lnodeT;
  if(sizeTi>0) delete inodeT;
  if(sizeQ>0) delete inodeQ;
}

void
eventQueue2::process(mObjs *mos) {

  mObj2D *mo1, *mo2;
  int ind,ind1,ind2,ne,neind,i,relind;
  REAL ct,cd;

  ct = inodeQ[0].T;

  if(ct==MAX) return;

  ind = inodeQ[0].ind;
  ind1 = inodeT[ind].ind;
  ind2 = inodeT[ind+1].ind;
  i = ind;
  if(inodeT[PA(ind)].ind == ind1) {
    ind1 = inodeT[ind+1].ind;
    ind2 = inodeT[ind].ind;
    i = ind+1;
  }

  mo1 = &(mos->mos[ind1/3]);
  mo2 = lnodeT[ind1];
  cd = mo1->distance2(*mo2,ct);
  inodeT[i].T = cd;

  ne = NE(i);
  relind = TQ(i,sizeTl);
  
  neind = inodeT[ne].ind;
  inodeQ[relind].T = meet(mo1,mo2,&(mos->mos[neind/3]),
			  lnodeT[neind],ct,inodeQ[relind].un+1,
			  &(inodeQ[relind].un));
  updateup(inodeQ,relind-sizeTl+2,sizeTl-1);

  i = PA(i);
  inodeT[i].ind = ind1;
  inodeT[i].T = cd;
  
  while(i!=0) {

    ne = NE(i);
    relind = TQ(i,sizeTl);
    
    neind = inodeT[ne].ind;
    inodeQ[relind].T = meet(mo1,mo2,&(mos->mos[neind/3]),
			    lnodeT[neind],ct,1,&(inodeQ[relind].un));
    updateup(inodeQ,relind-sizeTl+2,sizeTl-1);

    i = PA(i);
    if(inodeT[i].ind!=ind2) break;
    inodeT[i].ind = ind1;
    inodeT[i].T = cd;
  }
}    

void
eventQueue2::update(mObjs *mos, int index, short axis, REAL t) {

  mObj2D *lc;
  int ne,pa,relind,ind1,ind2,ind,newind,nnind;
  int i;

  i = 3*index+axis+sizeTl-1;
  
  lc = mos->mos[index].lcand[axis];
  ind = 3*index+axis;
  lnodeT[ind] = lc;
  if(lc==NULL) inodeT[i].T = MAX;
  else inodeT[i].T = mos->mos[index].distance2(*lc,t);

  while(i!=0) {

    ne = NE(i);
    pa = PA(i);
    relind = TQ(i,sizeTl);
    ind1 = inodeT[i].ind;
    ind2 = inodeT[ne].ind;
    inodeQ[relind].T = meet(&(mos->mos[ind1/3]),lnodeT[ind1],
			    &(mos->mos[ind2/3]),lnodeT[ind2],t,
			    1,&(inodeQ[relind].un));

    updateup(inodeQ,relind-sizeTl+2,sizeTl-1);
    if(inodeT[ne].T==MAX)
      newind = i;
    else {
      inodeT[ne].T = mos->mos[ind2/3].distance2(*lnodeT[ind2],t);      
      if(inodeT[i].T<=inodeT[ne].T) newind = i;
      else newind = ne;
    }
    inodeT[pa].T = inodeT[newind].T;
    nnind = inodeT[newind].ind;
    /*
    if(nnind!=ind) flag = 1;
    if(flag) 
      if(inodeT[pa].ind == nnind) break;*/
    inodeT[pa].ind = nnind;
    i = pa;
  }
}  

REAL 
eventQueue2::extractMinDist(void) {

  return inodeT[0].T;
}
   
REAL
eventQueue2::extractMinPair(mObjs *mos, mObj2D **mo1, mObj2D **mo2) {
  
  int ind=inodeT[0].ind;
  *mo1 = &(mos->mos[ind/3]);
  *mo2 = lnodeT[ind];
  return inodeT[0].T;
}

REAL
eventQueue2::extractMin() {

  return inodeQ[0].T;
}

REAL
eventQueue2::extractMin(mObjs *mos, mObj2D **mo1, mObj2D **mo2, short *ax) {

  int ind,ind1,ind2;
  if(inodeQ[0].T!=MAX) {

    ind = inodeQ[0].ind;
    ind1 = inodeT[ind].ind;
    ind2 = inodeT[ind+1].ind;
    if(inodeT[PA(ind)].ind == ind1) {
      *mo1 = &(mos->mos[ind2/3]);
      *mo2 = lnodeT[ind2];
      *ax = ind2%3;
    } else {
      *mo1 = &(mos->mos[ind1/3]);
      *mo2 = lnodeT[ind1];
      *ax = ind1%3;
    }
  }
  return inodeQ[0].T;
}

eventQueue3::eventQueue3(mObjs *mos) {

  sizei = 0;
  sizel = 0;
  init(mos);
}

void
eventQueue3::init(mObjs *mos) {

  b = &(mos->b);
  int nos;
  if(sizei>0) delete inode;
  if(sizel>0) delete lnode;
  
  nos = mos->nofObjs;
  if(nos<=0) return;
  
  sizel = nos;
  sizei = 2*nos-1;

  inode = (innerNode *) malloc(sizei*sizeof(innerNode));
  lnode = (short *) malloc(sizel*sizeof(short));
  
  int i,j;

  for(i=0,j=sizel-1;i<nos;i++,j++) {

    inode[j].ind = i;
    inode[j].T = mos->mos[i].hitBoundary(b,0,&lnode[i]);
  }

  buildup(inode,sizel);
}

REAL
eventQueue3::extractMin(void) {

  return inode[0].T;
}

REAL
eventQueue3::extractMin(mObjs *mos, mObj2D **mo, short* side) {

  int index = inode[0].ind;
  *mo = NULL;
  if(inode[0].T==MAX) return MAX;
  *mo = &(mos->mos[index]);
  *side = lnode[index];
  return inode[0].T;
}

void
eventQueue3::update(mObjs *mos, int index, REAL t) {
  
  inode[sizel+index-1].T = mos->mos[index].hitBoundary(b,t,&lnode[index]);
  updateup(inode,index,sizel);
}  

eventQueue3::~eventQueue3(void) {

  if(sizei>0) delete inode;
  if(sizel>0) delete lnode;
}
