/*
 *  pp.h 
 * 
 *  Definitions of constants and some inline functions
 *  for point operations.
 */

#ifndef _PP_H
#define _PP_H

#include <iostream.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#define W 100
#define L 100

#define X 0
#define Y 1
#define Z 2

#define X60 1 
#define X120 2
#define X180 3
#define X240 4
#define X300 5
#define X360 6  

#define UP 0
#define DOWN 1

#define XY 2
#define XYZ 3
#define MAX 100000

#define s60 0.866025403784439
#define c60 0.5
#define s120 0.866025403784439
#define c120 -0.5
#define c180 -1
#define s180 0
#define s240 -0.866025403784439
#define c240 -0.5
#define s300 -0.866025403784439
#define c300 0.5

#define S_3   1.732050807568878
#define S_3_2 0.866025403784439
#define S_3_3 0.577350269189626
#define S_3_6 0.288675134594813

#define LS(i) ((i<<1)+1)
#define RS(i) ((i<<1)+2)
#define PA(i) ((i-1)>>1)
#define NE(i) (((i)%2)?(i+1):(i-1))
#define TQ(i,l) ((l-2)+((i-1)>>1))

#define INCREASING 0
#define DECREASING 1

#define RED     0
#define BLACK   1

#define RIGHT 0
#define LEFT 1

#define EVE1 0
#define EVE2 1
#define EVE3 2
#define EVE4 3

#define FRAMES 100
#define epsilon 1.0001
#define SAVE 10

#define UPTYPE1 0
#define UPTYPE2 1

#define abs(x) ((x)>0)?(x):(-(x))
typedef float REAL;
typedef REAL P2[XY];
typedef REAL P3[XYZ];

struct Boundary {

  REAL left,right,bottom,top;
};

inline
void draw(Boundary bo, REAL ra) {

  REAL l=bo.left-ra, r=bo.right+ra;
  REAL b=bo.bottom-ra, t=bo.top+ra;
  glPointSize(3.0);
  glBegin(GL_LINE_LOOP);
  glVertex2f(l,b);
  glVertex2f(l,t);
  glVertex2f(r,t);
  glVertex2f(r,b);
  glEnd();
}

inline 
void print(P2 p) { 

  glBegin(GL_POINTS);
  glVertex2f(p[X],p[Y]);
  glEnd();
}

inline 
REAL order(P2 p, short axis)
{
  short ax = axis%6;

  switch(ax) {

  case X:
    return p[X]; 

  case X60:
    return c60*(p[X])+s60*(p[Y]);

  case X120:
    return c120*(p[X])+s120*(p[Y]);

  case X180:
    return -p[X];

  case X240:
    return c240*p[X] + s240*p[Y];
  
  case X300:
    return c300*p[X] + s300*p[Y];
  }
  return 0;
}

inline
REAL orderY(P2 p, short axis) {

  short ax = axis%6;

  switch(ax) {

  case X:
    return p[Y]; 

  case X60:
    return -s60*(p[X])+c60*(p[Y]);

  case X120:
    return -s120*(p[X])+c120*(p[Y]);

  case X180:
    return -p[Y];

  case X240:
    return -s240*p[X] + c240*p[Y];
  
  case X300:
    return -s300*p[X] + c300*p[Y];
  }
  return 0;
}

inline 
void print(P2 p,short axis)
{ cout<<"\n"<<"("<<p[X]<<","<<p[Y]<<")  "<<order(p,axis)<<"\n";}

inline 
e(P2 p1, P2 p2, short axis)  { return(order(p1,axis)==order(p2,axis));}

inline 
l(P2 p1, P2 p2, short axis)  { return(order(p1,axis)< order(p2,axis));}

inline 
le(P2 p1, P2 p2, short axis) { return(order(p1,axis)<=order(p2,axis));}

inline 
g(P2 p1, P2 p2, short axis)  { return(order(p1,axis)> order(p2,axis));}

inline 
ge(P2 p1, P2 p2, short axis) { return(order(p1,axis)>=order(p2,axis));}

inline
compare(P2 p1, P2 p2, short axis, short order)  
{return((order==INCREASING)?le(p1,p2,axis):ge(p1,p2,axis));}

inline void assign(P2 a, P2 b) {

  b[X] = a[X];
  b[Y] = a[Y];
}

inline REAL dist2(P2 a, P2 b) {

  REAL dx,dy;

  dx = (a)[X]-(b)[X]; 
  dy = (a)[Y]-(b)[Y];
  return(dx*dx+dy*dy); 
}

inline REAL norm(P2 a) {

  return sqrt(a[X]*a[X]+a[Y]*a[Y]);
}

inline
void normalize(P2 a) {

  REAL l = norm(a);
  if(l==0) return;
  (a)[X] /= l;
  (a)[Y] /= l;
}

inline 
REAL dotcross(P2 p0, P2 p1) {

  return p0[X]*p1[X]+p0[Y]*p1[Y];
}

void sort(P2*, int, int*, short, short);
void sort(REAL*, int*, int, short);
void sort2(REAL*, REAL*, int*, int, short);

void rotate(P2, P2, short axis);
void rotate(P2*,int, P2*, short);

inline
lessthan(REAL k11, REAL k12, REAL k21, REAL k22) {
  
  if(k11<k21) return 1;
  if(k11>k21) return 0;
  return(k12<k22);
}

inline
greaterthan(REAL k11, REAL k12, REAL k21, REAL k22) {
  
  if(k11>k21) return 1;
  if(k11<k21) return 0;
  return(k12>k22);
}

inline
lessthan(REAL *u, REAL *v, int i, int j) {
  
  return lessthan(u[i],v[i],u[j],v[j]);
}

inline
greaterthan(REAL *u, REAL *v, int i, int j) {
  
  return greaterthan(u[i],v[i],u[j],v[j]);
}

inline short
locate(REAL x, REAL left, REAL right) {

  return (x<left)?0:((x>right)?2:1);
}

inline short
locate(P2 p, Boundary b) {

  short l1 = locate(p[X],b.left,b.right);
  short l2 = locate(p[Y],b.bottom,b.top);
  return 3*l1+l2;
}

inline void
intersect(P2 p0,P2 p1,P2 p,short axis) {
  
  switch(axis) {
  case X:
    p[X] = S_3_2*(p1[Y]-p0[Y])+0.5*(p0[X]+p1[X]);
    p[Y] = S_3_6*(p1[X]-p0[X])+0.5*(p0[Y]+p1[Y]);
    break;
  case X60:
    p[X] = p0[X];
    p[Y] = S_3_3*(p0[X]-p1[X])+p1[Y];
    break;
  case X120:
    p[X] = p1[X];
    p[Y] = -S_3_3*(p1[X]-p0[X])+p0[Y];
    break;
  }
}

inline void
prune(P2 p0, P2 p, Boundary b) {

  short l = locate(p[X],b.left,b.right);
  REAL dx = p[X]-p0[X];
  REAL dy = p[Y]-p0[Y];
  
  if(l!=1) {
    
    if(l==0)  
      p[X] = b.left;
    else
      p[X] = b.right;
    p[Y] = p0[Y] + ((p[X]-p0[X])/dx)*dy;
  }
  
  l = locate(p[Y],b.bottom,b.top);
  if(l!=1) {
    
    if(l==0)
      p[Y] = b.bottom;
    else
      p[Y] = b.top;
    p[X] = p0[X] + ((p[Y]-p0[Y])/dy)*dx;
    
  }
}

inline
samepair(int a1, int a2, int b1, int b2) {

  if(a1==b1) return (a2==b2);
  return (a1==b2)&&(a2==b1);
}

inline
REAL
approx(REAL x, REAL y, REAL s, REAL r1, REAL r2) {
  
  if(x>s+r1) return x;
  if(x>s-r2) return y;
  if(y>s+r1) return y;
  return MAX;
}

#endif
